#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.08.20 00:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex, QSize
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QTreeView, QMenu, QStyledItemDelegate, QComboBox, \
    QCheckBox, QHeaderView, QHBoxLayout

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class OptionDict(QWidget):
    def __init__(self, window=None, parent_id: str = None, id: str = None, option: dict = None):
        """
        Settings dictionary items

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(OptionDict, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option
        self.title = ""
        self.real_time = False
        self.keys = {}
        self.items = []

        # init from option data
        if self.option is not None:
            if "label" in self.option:
                self.title = self.option["label"]
            if "value" in self.option:
                self.items = self.option["value"]
            if "keys" in self.option:
                self.keys = self.option["keys"]
            if "items" in self.option:
                self.items = list(self.option["items"])
            if "real_time" in self.option:
                self.real_time = self.option["real_time"]

        # setup dict model
        headers = list(self.keys.keys())

        self.list = OptionDictItems(self)
        max_height_delegate = MaxHeightDelegate(40, self.list)
        self.list.setItemDelegate(max_height_delegate)
        self.model = OptionDictModel(self.items, headers)
        self.model.dataChanged.connect(self.model.saveData)

        # append dict model
        self.list.setModel(self.model)

        # init layout
        self.init_layout()

        # update model list
        self.update()

    def init_layout(self):
        # add button
        self.add_btn = QPushButton(trans('action.add'), self)
        self.add_btn.setCursor(Qt.PointingHandCursor)
        self.add_btn.clicked.connect(self.add)
        self.add_btn.setAutoDefault(False)

        self.buttons = QHBoxLayout()
        self.buttons.addWidget(self.add_btn)
        self.buttons.setContentsMargins(0, 0, 0, 0)

        # layout
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addLayout(self.buttons)
        self.layout.addWidget(self.list)
        self.setLayout(self.layout)

    def add(self):
        """Add new empty item to settings dict"""
        # create empty dict
        empty = {header: '' for header in self.model.headers}
        count = self.model.rowCount()

        # add new item
        self.model.items.append(empty)
        self.model.updateData(self.model.items)

        # mark new item
        new_index = self.model.index(count, 0)
        self.list.setCurrentIndex(new_index)
        self.list.scrollTo(new_index)

    def edit_item(self, event):
        """
        Open item editor dialog

        :param event: Menu event
        """
        index = self.list.indexAt(event.pos())
        if index.isValid():
            idx = index.row()
            data = self.model.items[idx]
            id = self.parent_id + "." + self.id  # dictionary id
            self.window.ui.dialogs.open_dictionary_editor(id, self.option, data, idx)

    def update_item(self, idx, data):
        """
        Update item data in model by idx

        :param idx: Index of item
        :param data: Item data dict (key: value)
        """
        self.model.items[idx] = data
        self.model.updateData(self.model.items)

    def delete_item(self, event):
        """
        Delete item (show confirm)

        :param event: Menu event
        """
        index = self.list.indexAt(event.pos())
        if index.isValid():
            idx = index.row()
            self.window.controller.config.dictionary.delete_item(self, idx)

    def delete_item_execute(self, idx):
        """
        Delete item (execute)

        :param idx: Index of item
        """
        if len(self.model.items) > idx:
            self.model.items.pop(idx)
        self.model.removeRows(idx, 1)

        # update model list
        self.update()
        self.list.updateGeometries()
        self.items = self.model.items
        self.model.updateData(self.items)

    def remove(self):
        """
        Remove all items
        """
        self.model.removeRows(0, self.model.rowCount())

    def update(self):
        """
        Updates list of items
        """
        self.model.removeRows(0, self.model.rowCount())
        i = 0
        for item in list(self.items):
            self.model.insertRow(i)
            j = 0
            for k in self.keys:
                if k in item:
                    # add item to model list
                    self.model.setData(self.model.index(i, j), item[k])
                j += 1
            i += 1


class OptionDictItems(QTreeView):
    NAME = range(1)  # list of columns

    def __init__(self, owner=None):
        """
        Select menu

        :param window: main window
        :param id: input id
        """
        super(OptionDictItems, self).__init__(owner)
        self.parent = owner
        self.setIndentation(0)
        self.setHeaderHidden(False)

        header = self.header()
        header.setStretchLastSection(True)
        header.setSectionResizeMode(QHeaderView.Stretch)

        # setup fields
        keys = self.parent.keys
        idx = 0
        for key in keys:
            item = keys[key]
            if type(item) is dict:
                if "type" in item:
                    if item["type"] == "combo":
                        handler = ComboBoxDelegate(self, item["keys"])
                        self.setItemDelegateForColumn(idx, handler)
            elif type(item) is str:
                if item == "bool":
                    handler = CheckBoxDelegate(self)
                    self.setItemDelegateForColumn(idx, handler)
                elif item == "hidden":
                    continue
            idx += 1

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: context menu event
        """
        actions = {}
        actions['edit'] = QAction(QIcon(":/icons/edit.svg"), trans('action.edit'), self)
        actions['edit'].triggered.connect(
            lambda: self.parent.edit_item(event))
        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.delete'), self)
        actions['delete'].triggered.connect(
            lambda: self.parent.delete_item(event))
        menu = QMenu(self)
        menu.addAction(actions['edit'])
        menu.addAction(actions['delete'])

        # get index of item
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            menu.exec_(event.globalPos())


class OptionDictModel(QAbstractItemModel):
    def __init__(self, items, headers, parent=None):
        super(OptionDictModel, self).__init__(parent)
        self.items = items
        self.headers = headers
        self.checkbox_key = 'enabled'

    def headerData(self, section, orientation, role):
        """
        Header data
        :param section: Section
        :param orientation: Orientation
        :param role: Role
        :return: Header data
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.headers[section]
        return None

    def rowCount(self, parent=QModelIndex()):
        """
        Row count
        :param parent: Parent
        :return: Row count
        """
        return len(self.items) if not parent.isValid() else 0

    def columnCount(self, parent=QModelIndex()):
        """
        Column count
        :param parent: Parent
        :return: Column count
        """
        return len(self.headers) if not parent.isValid() else 0

    def data(self, index, role=Qt.DisplayRole):
        """
        Data
        :param index: Index
        :param role: Role
        :return: Data
        """
        if not index.isValid():
            return None
        if role == Qt.CheckStateRole and index.column() == 0 and self.headers[index.column()] == "enabled":
            value = self.items[index.row()].get('enabled', False)
            return Qt.Checked if value else Qt.Unchecked
        if role == Qt.EditRole and index.column() == 0 and self.headers[index.column()] == "enabled":
            return self.items[index.row()].get('enabled', False)
        if role == Qt.DisplayRole or role == Qt.EditRole:
            entry = self.items[index.row()]
            key = self.headers[index.column()]
            return entry.get(key, "")
        return None

    def setData(self, index, value, role=Qt.EditRole):
        """
        Set data
        :param index: Index
        :param value: Value
        :param role: Role
        :return: Set data
        """
        if index.isValid() and role == Qt.CheckStateRole and self.headers[index.column()] == "enabled":
            self.items[index.row()]['enabled'] = (value == Qt.Checked)
            self.dataChanged.emit(index, index, [Qt.CheckStateRole])
            return True

        if index.isValid() and role == Qt.EditRole:
            entry = self.items[index.row()]
            key = self.headers[index.column()]
            entry[key] = value
            self.dataChanged.emit(index, index, [Qt.EditRole])
            return True
        return False

    def flags(self, index):
        """
        Flags
        :param index: Index
        :return: Flags
        """
        if not index.isValid():
            return Qt.NoItemFlags
        if index.column() == 0 and self.headers[index.column()] == "enabled":
            return super(OptionDictModel, self).flags(index) | Qt.ItemIsUserCheckable | Qt.ItemIsEditable
        return super(OptionDictModel, self).flags(index) | Qt.ItemIsEditable

    def parent(self, index):
        """
        Parent
        :param index: Index
        :return: Parent
        """
        return QModelIndex()

    def index(self, row, column, parent=QModelIndex()):
        """
        Index
        :param row: Row
        :param column: Column
        :param parent: Parent
        :return: Index
        """
        if parent.isValid() or row >= len(self.items) or column >= len(self.headers):
            return QModelIndex()
        return self.createIndex(row, column)

    def saveData(self):
        """
        Save data
        """
        pass

    def updateData(self, data):
        """
        Update data
        :param new_data_list: New data list
        """
        self.beginResetModel()
        self.items = data
        self.endResetModel()


class CheckBoxDelegate(QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        editor = QCheckBox(parent)
        return editor

    def setEditorData(self, editor, index):
        checked = index.model().data(index, Qt.EditRole)
        if type(checked) is str:
            checked = checked == 'true'
        editor.setChecked(checked)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.isChecked(), Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class ComboBoxDelegate(QStyledItemDelegate):
    def __init__(self, parent=None, keys=None):
        super(ComboBoxDelegate, self).__init__(parent)
        self.keys = keys

    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        for item in self.keys:
            if isinstance(item, dict):
                for key, value in item.items():
                    editor.addItem(value, key)
            else:
                editor.addItem(item)
        return editor

    def setEditorData(self, editor, index):
        key = index.model().data(index, Qt.EditRole)
        if key is not None:
            index = editor.findData(key)
            if index >= 0:
                editor.setCurrentIndex(index)

    def setModelData(self, editor, model, index):
        key = editor.currentData()
        model.setData(index, key, Qt.EditRole)

    def displayText(self, value, locale):
        for item in self.keys:
            if isinstance(item, dict):
                if value in item:
                    return item[value]
        return value


class MaxHeightDelegate(QStyledItemDelegate):
    def __init__(self, max, parent=None):
        super().__init__(parent)
        self.maxHeight = max

    def sizeHint(self, option, index):
        original = super().sizeHint(option, index)
        limited = min(original.height(), self.maxHeight)
        return QSize(original.width(), limited)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 18:00:00                  #
# ================================================== #

import base64
from typing import Optional, Union, List, Dict, Any

from pygpt_net.core.bridge.context import MultimodalContext


class Audio:
    def __init__(self, window=None):
        """
        Audio input wrapper

        :param window: Window instance
        """
        self.window = window

    def build_content(
            self,
            content: Optional[Union[str, list]] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ) -> List[Dict[str, Any]]:
        """
        Build audio content from multimodal context

        :param content: previous content or input prompt
        :param multimodal_ctx: multimodal context
        :return: List of contents
        """
        if not isinstance(content, list):
            if content:
                content = [
                    {
                        "type": "text",
                        "text": str(content),
                    }
                ]
            else:
                content = []  # if empty input return empty list

        # abort if no audio input provided
        if not multimodal_ctx.is_audio_input:
            return content

        encoded = base64.b64encode(multimodal_ctx.audio_data).decode('utf-8')
        audio_format = multimodal_ctx.audio_format  # wav by default
        audio_data = {
            "type": "input_audio",
            "input_audio": {
                "data": encoded,
                "format": audio_format,
            }
        }
        content.append(audio_data)
        return content



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.17 02:00:00                  #
# ================================================== #

from .about import About
from .audio import Audio
from .config import Config
from .debug import Debug
from .donate import Donate
from .file import File
from .lang import Lang
from .plugins import Plugins
from .theme import Theme
from .tools import Tools
from .video import Video


class Menu:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window
        self.about = About(window)
        self.audio = Audio(window)
        self.config = Config(window)
        self.debug = Debug(window)
        self.donate = Donate(window)
        self.file = File(window)
        self.lang = Lang(window)
        self.plugins = Plugins(window)
        self.theme = Theme(window)
        self.tools = Tools(window)
        self.video = Video(window)

    def setup(self):
        """Setup all menus"""
        self.window.ui.menu = {}
        self.file.setup()
        self.plugins.setup()
        self.audio.setup()
        self.video.setup()
        self.config.setup()

    def post_setup(self):
        """Post setup menus"""
        # tools menu
        self.tools.setup()
        self.about.setup()

        # debug menu
        show = self.window.core.config.get('debug')
        self.debug.setup()
        self.window.ui.menu['menu.debug'].menuAction().setVisible(show)

        self.donate.setup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional

from PySide6.QtWidgets import QApplication

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.utils import trans
from .summarizer import Summarizer


class Common:
    def __init__(self, window=None):
        """
        Common ctx controller

        :param window: Window instance
        """
        self.window = window
        self.summarizer = Summarizer(window)

    def update_label_by_current(self):
        """Update ctx label from current ctx"""
        mode = self.window.core.ctx.get_mode()

        # if no ctx mode then use current mode
        if mode is None:
            mode = self.window.core.config.get('mode')

        label = trans('mode.' + mode)

        # append assistant name to ctx name label
        if mode == 'assistant':
            id = self.window.core.ctx.get_assistant()
            assistant = self.window.core.assistants.get_by_id(id)
            if assistant is not None:
                # get ctx assistant
                label += ' (' + assistant.name + ')'
            else:
                # get current assistant
                id = self.window.core.config.get('assistant')
                assistant = self.window.core.assistants.get_by_id(id)
                if assistant is not None:
                    label += ' (' + assistant.name + ')'

        # update ctx label
        self.window.controller.ui.update_ctx_label(label)

    def update_label(
            self,
            mode: str,
            assistant_id: Optional[str] = None
    ):
        """
        Update ctx label

        :param mode: Mode
        :param assistant_id: Assistant id
        """
        if mode is None:
            return
        label = trans('mode.' + mode)
        if mode == 'assistant' and assistant_id is not None:
            assistant = self.window.core.assistants.get_by_id(assistant_id)
            if assistant is not None:
                label += ' (' + assistant.name + ')'

        # update ctx label
        self.window.controller.ui.update_ctx_label(label)

    def duplicate(self, meta_id: int):
        """
        Duplicate context by meta id

        :param meta_id: context id
        """
        new_id = self.window.core.ctx.duplicate(meta_id)
        if new_id is not None:
            self.window.core.attachments.context.duplicate(meta_id, new_id)  # copy attachments
            self.window.update_status(
                "Context duplicated, new ctx id: {}".format(new_id)
            )
            self.window.controller.ctx.update()

    def dismiss_rename(self):
        """Dismiss rename dialog"""
        self.window.ui.dialog['rename'].close()

    def focus_chat(self):
        """Focus chat"""
        tabs = self.window.ui.layout.get_active_tabs()
        if self.window.controller.ui.tabs.get_current_type() != Tab.TAB_CHAT:
            idx = self.window.core.tabs.get_min_idx_by_type(Tab.TAB_CHAT)
            if idx is not None:
                tabs.setCurrentIndex(idx)

    def restore_display_filter(self):
        """Restore display filter"""
        self.window.ui.nodes['filter.ctx.radio.all'].setChecked(False)
        self.window.ui.nodes['filter.ctx.radio.pinned'].setChecked(False)
        self.window.ui.nodes['filter.ctx.radio.indexed'].setChecked(False)

        if self.window.core.config.has('ctx.records.filter'):
            filter = self.window.core.config.get('ctx.records.filter')
            self.toggle_display_filter(filter)

            if filter == 'pinned':
                self.window.ui.nodes['filter.ctx.radio.pinned'].setChecked(True)
            elif filter == 'indexed':
                self.window.ui.nodes['filter.ctx.radio.indexed'].setChecked(True)
            else:
                self.window.ui.nodes['filter.ctx.radio.all'].setChecked(True)
        else:
            self.window.ui.nodes['filter.ctx.radio.all'].setChecked(True)
            self.toggle_display_filter('all')

        # restore filters labels
        self.restore_filters_labels()

    def restore_filters_labels(self):
        """Restore filters labels"""
        labels = self.window.core.config.get('ctx.records.filter.labels')
        if labels is not None:
            self.window.core.ctx.filters_labels = labels
            self.window.ui.nodes['filter.ctx.labels'].restore(labels)

    def toggle_display_filter(self, filter: str):
        """
        Toggle display filter

        :param filter: Filter
        """
        filters = {}
        if filter == 'pinned':
            filters['is_important'] = {
                "mode": "=",
                "value": 1,
            }
        elif filter == 'indexed':
            filters['indexed_ts'] = {
                "mode": ">",
                "value": 0,
            }

        self.window.core.config.set("ctx.records.filter", filter)
        self.window.core.ctx.clear_tmp_meta()
        self.window.core.ctx.set_display_filters(filters)
        self.window.controller.ctx.update()

    def copy_id(self, id: int):
        """
        Copy id into clipboard and to iinput

        :param id: context list idx
        """
        value = "@" + str(id)
        self.window.controller.chat.common.append_to_input(value, separator=" ")
        QApplication.clipboard().setText(value)

    def reset(
            self,
            meta_id: int,
            force: bool = False
    ):
        """
        Reset by meta id

        :param meta_id: context id
        :param force: True to force reset
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.reset_meta',
                id=meta_id,
                msg=trans('ctx.reset_meta.confirm'),
            )
            return
        self.window.core.ctx.reset_meta(meta_id)
        self.window.core.attachments.context.reset_by_meta_id(meta_id, delete_files=True)
        self.window.controller.ctx.load(meta_id)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.05 23:00:00                  #
# ================================================== #

import os

import markdown
from mdx_math import MathExtension
from bs4 import BeautifulSoup
from bs4.element import NavigableString
from pygpt_net.utils import trans


class Parser:

    def __init__(self, window=None):
        """
        Markdown parser core

        :param window: Window instance
        """
        self.window = window
        self.md = None
        self.code_blocks = {}
        self.block_idx = 1

    def init(self):
        """Initialize markdown parser"""
        if self.md is None:
            self.md = markdown.Markdown(extensions=[
                'fenced_code',
                'tables',
                'sane_lists',
                MathExtension(enable_dollar_delimiter=True)  # math formulas
            ])
            # disable code blocks parsing (without ` and ```)
            if 'code' in self.md.parser.blockprocessors:
                self.md.parser.blockprocessors.deregister('code')

    def reset(self):
        """
        Reset parser
        """
        self.code_blocks = {}
        self.block_idx = 1

    def get_code_blocks(self):
        """
        Get code blocks

        :return: code blocks
        """
        return self.code_blocks

    def prepare_paths(self, text: str) -> str:
        """
        Prepare paths in markdown text

        :param text: markdown text
        :return: markdown text with prepared paths
        """
        # Replace sandbox paths with file paths
        return text.replace("](sandbox:", "](file://")

    def parse(self, text: str) -> str:
        """
        Parse markdown text

        :param text: markdown text
        :return: html formatted text
        """
        self.init()
        try:
            text = self.prepare_paths(text.strip())
            html = self.md.convert(text)
            soup = BeautifulSoup(html, 'html.parser')
            self.strip_whitespace_lists(soup)  # strip whitespace from codeblocks
            # if self.window.core.config.get("ctx.convert_lists"):
            #   self.convert_lists_to_paragraphs(soup)  # convert lists to paragraphs, DISABLED: 2024-11-03
            self.strip_whitespace_codeblocks(soup)  # strip whitespace from codeblocks
            self.highlight_code_blocks(soup)  # parse code blocks
            self.format_images(soup)  # add width to img tags
            return str(soup)
        except Exception as e:
            pass
        return text

    def parse_code(self, text: str) -> str:
        """
        Parse code markdown text

        :param text: markdown text
        :return: html formatted text
        """
        self.init()
        try:
            soup = BeautifulSoup(self.md.convert(text.strip()), 'html.parser')
            self.strip_whitespace_lists(soup)  # strip whitespace from codeblocks
            self.strip_whitespace_codeblocks(soup)  # strip whitespace from codeblocks
            self.highlight_code_blocks(soup)  # parse code blocks
            return str(soup)
        except Exception as e:
            pass
        return text

    def to_plain_text(self, html: str) -> str:
        """
        Convert markdown to plain text

        :param html: html text
        :return: plain text
        """
        soup = BeautifulSoup(html, 'html.parser')
        text = soup.get_text()
        return text

    def strip_whitespace_codeblocks(self, soup):
        """
        Strip whitespace from codeblocks

        :param soup: BeautifulSoup instance
        """
        for el in soup.find_all('code'):
            el.string = el.string.strip()

    def strip_whitespace_lists(self, soup):
        """
        Strip whitespace from lists

        :param soup: BeautifulSoup instance
        """
        for li in soup.find_all('li'):
            for item in li.contents:
                if isinstance(item, NavigableString) and item.strip() == '':
                    item.replace_with('')

    def parse_code_blocks(self, soup):
        """
        Add copy code button to code blocks (without syntax highlighting)

        :param soup: BeautifulSoup instance
        """
        for el in soup.find_all('pre'):
            content = el.get_text(strip=True)
            self.code_blocks[self.block_idx] = content

            header = soup.new_tag('p', **{'class': "code-header-wrapper"})
            link_wrapper = soup.new_tag('span')
            a = soup.new_tag('a', href=f'extra-code-copy:{self.block_idx}')  # extra action link
            a['class'] = "code-header-copy"
            a.string = trans('ctx.extra.copy_code')

            code = el.find('code')
            language = code['class'][0] if code.has_attr('class') else ''
            if language:
                lang = soup.new_tag('span', **{'class': "code-header-lang"})
                lang['class'] = "code-header-lang"
                lang.string = language.replace('language-', '') + "   "
                link_wrapper.append(lang)

            link_wrapper.append(a)
            header.append(link_wrapper)
            wrapper = soup.new_tag('div', **{'class': "code-wrapper"})
            wrapper.append(header)

            new_code = soup.new_tag('code')
            new_pre = soup.new_tag('pre')
            new_code.string = content
            new_pre.append(new_code)
            wrapper.append(new_pre)
            el.replace_with(wrapper)

            self.block_idx += 1

    def highlight_code_blocks(self, soup):
        """
        Add copy code button to code blocks (with syntax highlighting)

        :param soup: BeautifulSoup instance
        """
        icon_path = os.path.join(self.window.core.config.get_app_path(), "data", "icons", "chat", "copy.png")
        style = self.window.core.config.get("render.code_syntax")
        if style is None or style == "":
            style = "default"

        for el in soup.find_all('pre'):
            content = el.text
            self.code_blocks[self.block_idx] = content

            # ignore empty
            if content.strip() == "":
                continue

            header = soup.new_tag('p', **{'class': "code-header-wrapper"})
            link_wrapper = soup.new_tag('span')
            a = soup.new_tag('a', href=f'empty:{self.block_idx}')  # extra action link
            a['class'] = "code-header-copy"
            a.string = trans('ctx.extra.copy_code')

            icon = soup.new_tag('img', src=icon_path, **{'class': "action-img"})
            a.insert(0, icon)

            # Get the class of <code> to determine the language (if available)
            code = el.find('code')
            language = code['class'][0] if (code and code.has_attr('class') and 'language-' in code['class'][0]) else ''

            lang_span = soup.new_tag('span', **{'class': "code-header-lang"})
            lang_span['class'] = "code-header-lang"
            if language:  # Strip 'language-' to get the actual language
                language = language.replace('language-', '')
                lang_span.string = language + "   "
            else:
                lang_span.string = "code "

            link_wrapper.append(lang_span)
            link_wrapper.append(a)
            header.append(link_wrapper)

            # Create wrapper to hold both the header and the code block
            wrapper = soup.new_tag('div', **{'class': "code-wrapper highlight"})
            wrapper.append(header)

            # DEPRECATED: Use Pygments to highlight the code block
            """
            # Use Pygments to highlight the code block
            if language:
                try:
                    lexer = get_lexer_by_name(language)
                except Exception:
                    lexer = get_lexer_by_name("sh")
            else:
                lexer = get_lexer_by_name("sh")

            
            # formatter = HtmlFormatter(style=style, cssclass='source', lineanchors='line')
            # highlighted_code = highlight(content, lexer, formatter)
            # Replace the original code block with the highlighted version
            # new_code = BeautifulSoup(highlighted_code, 'html.parser')
            # wrapper.append(new_code)
            # el.replace_with(wrapper)
            """

            pre = soup.new_tag('pre')
            code = soup.new_tag('code')
            if language:
                code['class'] = 'language-' + language
            code.string = content
            pre.append(code)
            wrapper.append(pre)
            el.replace_with(wrapper)

            self.block_idx += 1

    def convert_lists_to_paragraphs(self, soup):
        """
        Convert lists to paragraphs

        :param soup: BeautifulSoup instance
        """
        for ul in soup.find_all('ul'):
            self.convert_list(soup, ul, ordered=False)
        for ol in soup.find_all('ol'):
            self.convert_list(soup, ol, ordered=True)

        for element in soup.find_all(['ul', 'ol']):
            element.decompose()

    def convert_list(self, soup, list_element, ordered=False):
        """
        Convert list to paragraphs

        :param soup: BeautifulSoup instance
        :param list_element: Element to convert
        :param ordered: Is ordered list
        """
        list_items = list_element.find_all('li')
        for index, li in enumerate(list_items, start=1):
            p = soup.new_tag('p')
            p['class'] = "list"
            prefix = f"{index}. " if ordered else "- "
            p.string = f"{prefix}{li.get_text().strip()}"
            list_element.insert_before(p)

    def format_images(self, soup):
        """
        Add width to img tags

        :param soup: BeautifulSoup instance
        """
        images = soup.find_all('img')
        for index, img in enumerate(images):
            img['width'] = "400"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 17:00:00                  #
# ================================================== #

from pygpt_net.core.events import (
    BaseEvent,
    KernelEvent,
    ControlEvent,
    AppEvent,
    RenderEvent,
)


class Dispatcher:
    def __init__(self, window=None):
        """
        Event dispatcher

        :param window: Window instance
        """
        self.window = window
        self.nolog_events = ["system.prompt", "render.stream.append"]
        self.call_id = 0

    def dispatch(
            self,
            event: BaseEvent,
            all: bool = False
    ) -> (list, BaseEvent):
        """
        Dispatch event

        :param event: event to dispatch
        :param all: true if dispatch to all plugins (enabled or not)
        :return: list of affected plugins ids and event object
        """
        if not isinstance(event, RenderEvent):
            event.call_id = self.call_id

        if self.is_log(event):
            self.window.core.debug.info("[event] Dispatch begin: " + event.full_name + " (" + str(event.call_id) + ")")
            if self.window.core.debug.enabled():
                self.window.core.debug.debug("[event] Before handle: " + str(event))

        if event.stop:
            self.window.core.debug.info("[event] Skipping... (stopped): " + event.name)
            return

        # kernel events
        if isinstance(event, KernelEvent):
            # dispatch event to kernel controller
            if not event.name in [  # those events are sent by kernel
                KernelEvent.INIT,
                KernelEvent.RESTART,
                KernelEvent.STOP,
                KernelEvent.TERMINATE,
            ]:
                self.window.controller.kernel.handle(event)
            if self.is_log(event):
                self.window.core.debug.info("[event] Dispatch end: " + event.full_name + " (" + str(event.call_id) + ")")
            self.call_id += 1
            if not event.name in [
                KernelEvent.INIT,
                KernelEvent.RESTART,
                KernelEvent.STOP,
                KernelEvent.TERMINATE,
            ]:
                return [], event # kernel events finish here

        # render events
        elif isinstance(event, RenderEvent):
            # dispatch event to render controller
            self.window.controller.chat.render.handle(event)
            if self.is_log(event):
                self.window.core.debug.info("[event] Dispatch end: " + event.full_name + " (" + str(event.call_id) + ")")
            self.call_id += 1
            return [], event

        # accessibility events
        if isinstance(event, ControlEvent) or isinstance(event, AppEvent):
            # dispatch event to accessibility controller
            self.window.controller.access.handle(event)

        # dispatch event to plugins
        affected = []
        for id in list(self.window.core.plugins.plugins.keys()):
            if self.window.controller.plugins.is_enabled(id) or all:
                if event.stop:
                    self.window.core.debug.info("[event] Skipping... (stopped):  " + event.name)
                    break
                if self.window.core.debug.enabled() and self.is_log(event):
                    self.window.core.debug.debug("[event] Apply [{}] to plugin: ".format(event.name) + id)
                self.apply(id, event)
                affected.append(id)

        if self.is_log(event):
            if self.window.core.debug.enabled():
                self.window.core.debug.debug("[event] After handle: " + str(event))
            self.window.core.debug.info("[event] Dispatch end: " + event.full_name + " (" + str(event.call_id) + ")")

        self.call_id += 1
        return affected, event

    def apply(
            self,
            id: str,
            event: BaseEvent
    ):
        """
        Handle event in plugin with provided id

        :param id: plugin id
        :param event: event object
        """
        if id in self.window.core.plugins.plugins:
            try:
                self.window.core.plugins.plugins[id].handle(event)
            except AttributeError:
                pass

    def is_log(self, event: BaseEvent) -> bool:
        """
        Check if event can be logged

        :param event: event object
        :return: true if can be logged
        """
        if event.name in self.nolog_events:
            return False
        if not self.is_log_display():
            return False
        data = event.data
        if data is not None and "silent" in data and data["silent"]:
            return False
        return True

    def is_log_display(self) -> bool:
        """
        Check if logging enabled

        :return: True logging enabled
        """
        return self.window.core.config.get("log.events", False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.12 01:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import Slot

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .docker import Docker
from .builder import Builder
from .ipython import LocalKernel
from .ipython import DockerKernel
from .output import Output
from .runner import Runner
from .worker import Worker

from pygpt_net.utils import trans


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_code_interpreter"
        self.name = "Code Interpreter (v2)"
        self.description = "Provides Python/HTML/JS code execution"
        self.prefix = "Code"
        self.type = [
            'interpreter',
        ]
        self.order = 100
        self.allowed_cmds = [
            #"ipython_execute_new",
            "ipython_execute",
            "ipython_kernel_restart",
            "code_execute",
            "code_execute_file",
            "code_execute_all",
            "get_python_output",
            "get_python_input",
            "clear_python_output",
            "render_html_output",
            "get_html_output",
        ]
        self.use_locale = True
        self.docker = Docker(self)
        self.runner = Runner(self)
        self.ipython_docker = DockerKernel(self)
        self.ipython_local = LocalKernel(self)
        self.builder = Builder(self)
        self.output = Output(self)
        self.worker = None
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx
        silent = data.get("silent", False)

        if name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
                silent,
            )

        elif name == Event.TOOL_OUTPUT_RENDER:
            if data['tool'] == self.id:
                data['html'] = self.output.handle(ctx, data['content'])

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        # get current working directory
        legacy_data = self.window.core.config.get_user_dir('data')
        ipython_data = os.path.join(legacy_data, 'ipython')

        for item in self.allowed_cmds:
            if self.has_cmd(item):
                cmd = self.get_cmd(item)
                if item in ["ipython_execute", "ipython_execute_new"]:
                    if self.get_option_value("sandbox_ipython"):
                        cmd["instruction"] += ("\nIPython works in Docker container. Directory /data is the container's workdir - "
                                           "directory is mapped as volume in host machine to: {}").format(ipython_data)
                elif item in ["code_execute", "code_execute_file", "code_execute_all"]:
                    if self.get_option_value("sandbox_docker"):
                        cmd["instruction"] += ("\nPython works in Docker container. Directory /data is the container's workdir - "
                                           "directory is mapped as volume in host machine to: {}").format(legacy_data)
                data['cmd'].append(cmd)  # append command


    def cmd(self, ctx: CtxItem, cmds: list, silent: bool = False):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        :param silent: silent mode
        """
        is_cmd = False
        force = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True
                if "force" in item and item["force"]:
                    force = True  # call from tool

        if not is_cmd:
            return

        # ipython
        if self.get_option_value("sandbox_ipython"):
            ipython_commands = [
                "ipython_execute_new",
                "ipython_execute",
                "ipython_kernel_restart",
            ]
            if any(x in [x["cmd"] for x in my_commands] for x in ipython_commands):
                # check for Docker installed
                if not self.get_interpreter().is_docker_installed():
                    # snap version
                    if self.window.core.platforms.is_snap():
                        self.error(trans('ipython.docker.install.snap'))
                        self.window.update_status(trans('ipython.docker.install.snap'))
                    # other versions
                    else:
                        self.error(trans('ipython.docker.install'))
                        self.window.update_status(trans('ipython.docker.install'))
                    return
                # check if image exists
                if not self.get_interpreter().is_image():
                    self.error(trans('ipython.image.build'))
                    self.window.update_status(trans('ipython.docker.build.start'))
                    self.builder.build_image()
                    return

        # legacy python
        if self.get_option_value("sandbox_docker"):
            sandbox_commands = [
                "code_execute",
                "code_execute_all",
                "code_execute_file",
            ]
            if any(x in [x["cmd"] for x in my_commands] for x in sandbox_commands):
                # check for Docker installed
                if not self.docker.is_docker_installed():
                    # snap version
                    if self.window.core.platforms.is_snap():
                        self.error(trans('docker.install.snap'))
                        self.window.update_status(trans('docker.install.snap'))
                    # other versions
                    else:
                        self.error(trans('docker.install'))
                        self.window.update_status(trans('docker.install'))
                    return
                # check if image exists
                if not self.docker.is_image():
                    self.error(trans('docker.image.build'))
                    self.window.update_status(trans('docker.build.start'))
                    self.docker.build()
                    return

        # set state: busy
        if not silent:
            self.cmd_prepare(ctx, my_commands)

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.cmds = my_commands
            worker.ctx = ctx

            # connect signals
            worker.signals.output.connect(self.handle_interpreter_output)
            worker.signals.clear.connect(self.handle_interpreter_clear)
            worker.signals.html_output.connect(self.handle_html_output)
            worker.signals.ipython_output.connect(self.handle_ipython_output)
            self.get_interpreter().attach_signals(worker.signals)
            self.runner.attach_signals(worker.signals)

            if not self.is_async(ctx) and not force:
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)

    @Slot(object)
    def handle_ipython_output(self, data):
        """
        Handle IPython output

        :param data: output data
        """
        if not self.get_option_value("attach_output"):
            return
        # if self.is_threaded():
            # return
        # print(data)
        cleaned_data = self.get_interpreter().remove_ansi(data)
        self.window.tools.get("interpreter").append_output(cleaned_data)
        if self.window.tools.get("interpreter").opened:
            self.window.update_status("")

    @Slot(object, str)
    def handle_interpreter_output(self, data, type):
        """
        Handle interpreter output

        :param data: output data
        :param type: output type
        """
        if not self.get_option_value("attach_output"):
            return
        self.window.tools.get("interpreter").append_output(data, type)

    @Slot()
    def handle_interpreter_clear(self):
        """Handle interpreter clear"""
        self.window.tools.get("interpreter").clear_all()

    @Slot(object)
    def handle_html_output(self, data):
        """
        Handle HTML/JS canvas output

        :param data: HTML/JS code
        """
        self.window.tools.get("html_canvas").set_output(data)
        self.window.tools.get("html_canvas").auto_open()
        
    def get_interpreter(self):
        """
        Get interpreter

        :return: interpreter
        """
        if self.get_option_value("sandbox_ipython"):
            return self.ipython_docker
        else:
            return self.ipython_local



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtCore import Qt, QObject, Signal, Slot
from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QTextEdit, QApplication, QVBoxLayout, QLabel, QCheckBox, QPushButton, QWidget, QSplitter, \
    QHBoxLayout

from pygpt_net.ui.widget.textarea.editor import BaseCodeEditor

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ToolWidget:
    def __init__(self, window=None, tool=None):
        """
        Python Code Interpreter

        :param window: Window instance
        :param tool: Tool instance
        """
        self.window = window # window instance
        self.tool = tool  # tool instance
        self.history = None # history
        self.input = None  # input
        self.output = None  # output
        self.checkbox_all = None  # all checkbox
        self.checkbox_auto_clear = None  # auto clear checkbox
        self.checkbox_ipython = None  # IPython checkbox
        self.btn_send = None  # send button
        self.btn_clear = None  # clear button
        self.label_output = None  # output label
        self.label_history = None  # history label

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.output.set_tab(tab)
        self.input.set_tab(tab)

    def setup(self, all: bool = True) -> QVBoxLayout:
        """
        Setup widget body

        :param all: with all widgets
        :return: QVBoxLayout
        """
        self.output = PythonOutput(self.window, self.tool)
        self.output.setReadOnly(True)

        if all:
            self.history = PythonOutput(self.window, self.tool)
            self.history.textChanged.connect(
                lambda: self.tool.store_history(self)
            )
            self.history.setReadOnly(False)
            self.history.excluded_copy_to = ["interpreter_edit"]

        self.label_output = QLabel(trans("interpreter.edit_label.output"))
        self.label_history = QLabel(trans("interpreter.edit_label.edit"))

        if all:
            self.checkbox_all = QCheckBox(trans("interpreter.all"))
            self.checkbox_all.setChecked(True)
            self.checkbox_all.clicked.connect(
                lambda: self.tool.toggle_all(self)
            )

        self.checkbox_auto_clear = QCheckBox(trans("interpreter.auto_clear"))
        self.checkbox_auto_clear.setChecked(False)
        self.checkbox_auto_clear.clicked.connect(
            lambda: self.tool.toggle_auto_clear(self)
        )

        self.checkbox_ipython = QCheckBox("IPython")
        self.checkbox_ipython.setChecked(True)
        self.checkbox_ipython.clicked.connect(
            lambda: self.tool.toggle_ipython(self)
        )

        self.btn_clear = QPushButton(trans("interpreter.btn.clear"))
        self.btn_clear.clicked.connect(
            lambda: self.tool.clear(self)
        )

        self.btn_send = QPushButton(trans("interpreter.btn.send"))
        self.btn_send.clicked.connect(
            lambda: self.tool.send_input(self)
        )

        self.input = PythonInput(self.window, self.tool, self)
        self.input.setPlaceholderText(trans("interpreter.input.placeholder"))
        self.input.excluded_copy_to = ["interpreter_input"]

        left_layout = QVBoxLayout()
        left_layout.addWidget(self.label_output)
        left_layout.addWidget(self.output)
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_widget = QWidget()
        left_widget.setLayout(left_layout)

        self.window.ui.splitters['interpreter.columns'] = QSplitter(Qt.Horizontal)
        self.window.ui.splitters['interpreter.columns'].addWidget(left_widget)

        if all:
            right_layout = QVBoxLayout()
            right_layout.addWidget(self.label_history)
            right_layout.addWidget(self.history)
            right_layout.setContentsMargins(0, 0, 0, 0)
            right_widget = QWidget()
            right_widget.setLayout(right_layout)
            right_widget.setMinimumWidth(300)
            self.window.ui.splitters['interpreter.columns'].addWidget(right_widget)

        bottom_layout = QHBoxLayout()
        bottom_layout.addWidget(self.btn_clear)
        bottom_layout.addWidget(self.checkbox_ipython)
        bottom_layout.addWidget(self.checkbox_auto_clear)
        bottom_layout.addStretch()
        if all:
            bottom_layout.addWidget(self.checkbox_all)
        bottom_layout.addWidget(self.btn_send)

        edit_layout = QVBoxLayout()
        edit_layout.addWidget(self.window.ui.splitters['interpreter.columns'])
        edit_layout.setContentsMargins(0, 0, 0, 0)
        edit_widget = QWidget()
        edit_widget.setLayout(edit_layout)

        self.window.ui.splitters['interpreter'] = QSplitter(Qt.Vertical)
        self.window.ui.splitters['interpreter'].addWidget(edit_widget)
        self.window.ui.splitters['interpreter'].addWidget(self.input)
        self.window.ui.splitters['interpreter'].setStretchFactor(0, 4)
        self.window.ui.splitters['interpreter'].setStretchFactor(1, 1)

        # connect signals
        self.tool.signals.update.connect(self.set_output)
        self.tool.signals.update_history.connect(self.set_history)
        self.tool.signals.clear_history.connect(self.clear_history)
        self.tool.signals.clear_output.connect(self.clear_output)
        self.tool.signals.focus_input.connect(self.set_focus)
        self.tool.signals.append_input.connect(self.append_to_input)
        self.tool.signals.update_input.connect(self.set_input)
        self.tool.signals.set_checkbox_all.connect(self.set_checkbox_all)
        self.tool.signals.set_checkbox_auto_clear.connect(self.set_checkbox_auto_clear)
        self.tool.signals.set_checkbox_ipython.connect(self.set_checkbox_ipython)
        self.tool.signals.toggle_all_visible.connect(self.toggle_all_visible)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.splitters['interpreter'])
        layout.addLayout(bottom_layout)
        return layout

    @Slot(str, str)
    def set_output(self, output: str, type="stdout"):
        """
        Set output content

        :param output: Content
        :param type: Output type
        """
        area = self.output
        if type == "stdin":
            data = ">> " + str(output)
        else:
            data = str(output)
        cur = area.textCursor()
        cur.movePosition(QTextCursor.End)
        s = data + "\n"
        while s:
            head, sep, s = s.partition("\n")
            cur.insertText(head)
            if sep:  # New line if LF
                cur.insertText("\n")
        area.setTextCursor(cur)

    @Slot()
    def set_history(self, data: str):
        """
        Set history

        :param data: history data
        """
        if not self.history:
            return
        self.history.setPlainText(data)
        cur = self.history.textCursor()
        cur.movePosition(QTextCursor.End)
        self.history.setTextCursor(cur)

    @Slot()
    def append_to_input(self, data: str):
        """
        Append data to input

        :param data: Data
        """
        current = self.input.toPlainText()
        if current:
            current += "\n"
        current += data
        self.input.setPlainText(current)

    @Slot(str)
    def set_input(self, data: str):
        """
        Set input

        :param data: Data
        """
        self.input.setPlainText(data)

    @Slot()
    def clear_history(self):
        """Clear input"""
        if not self.history:
            return
        self.history.clear()

    @Slot()
    def clear_output(self):
        """Clear output"""
        self.output.clear()

    @Slot()
    def set_focus(self):
        """Set focus to input"""
        self.input.setFocus()

    @Slot(bool)
    def set_checkbox_all(self, value: bool):
        """
        Set checkbox all

        :param value: Value
        """
        if not self.checkbox_all:
            return
        self.checkbox_all.setChecked(value)

    @Slot(bool)
    def set_checkbox_auto_clear(self, value: bool):
        """
        Set checkbox auto clear

        :param value: Value
        """
        self.checkbox_auto_clear.setChecked(value)

    @Slot(bool)
    def set_checkbox_ipython(self, value: bool):
        """
        Set checkbox IPython

        :param value: Value
        """
        self.checkbox_ipython.setChecked(value)

    @Slot(bool)
    def toggle_all_visible(self, value: bool):
        """
        Toggle all visible

        :param value: Value
        """
        if not self.checkbox_all:
            return
        if value:
            self.checkbox_all.setVisible(True)
        else:
            self.checkbox_all.setVisible(False)


class PythonInput(QTextEdit):
    def __init__(self, window=None, tool=None, widget=None):
        """
        Python interpreter input

        :param window: main window
        """
        super(PythonInput, self).__init__(window)
        self.window = window
        self.tool = tool
        self.widget = widget
        self.setAcceptRichText(False)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.setProperty('class', 'interpreter-input')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)
        self.textChanged.connect(
            lambda: self.tool.update_input()
        )
        self.setFocus()
        self.tab = None
        self.installEventFilter(self)

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def update_stylesheet(self, data: str):
        """
        Update stylesheet

        :param data: stylesheet CSS
        """
        self.setStyleSheet(self.default_stylesheet + data)

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        super(PythonInput, self).keyPressEvent(event)
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            mode = self.window.core.config.get('send_mode')
            if mode > 0:  # Enter or Shift + Enter
                if mode == 2:  # Shift + Enter
                    modifiers = QApplication.keyboardModifiers()
                    if modifiers == QtCore.Qt.ShiftModifier:
                        self.tool.send_input(self.widget)
                else:  # Enter
                    modifiers = QApplication.keyboardModifiers()
                    if modifiers != QtCore.Qt.ShiftModifier:
                        self.tool.send_input(self.widget)
                self.setFocus()

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            size_str = f"{self.value}px"
            self.update_stylesheet(f"font-size: {size_str};")
            event.accept()
        else:
            super(PythonInput, self).wheelEvent(event)


class PythonOutput(BaseCodeEditor):
    def __init__(self, window=None, tool=None):
        """
        Python interpreter output

        :param window: main window
        """
        super(PythonOutput, self).__init__(window)
        self.window = window
        self.tool = tool
        self.setReadOnly(True)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.setProperty('class', 'interpreter-output')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)
        self.tab = None
        self.installEventFilter(self)

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def clear_content(self):
        """Clear content"""
        super(PythonOutput, self).clear_content()
        self.tool.save_output()


class ToolSignals(QObject):
    update = Signal(str, str)  # content, type
    update_history = Signal(str)  # content
    update_input = Signal(str) # content
    append_input = Signal(str) # content
    reload = Signal(str)  # path
    clear_history = Signal()
    clear_output = Signal()
    focus_input = Signal()
    set_checkbox_all = Signal(bool)
    set_checkbox_auto_clear = Signal(bool)
    set_checkbox_ipython = Signal(bool)
    toggle_all_visible = Signal(bool)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.26 23:00:00                  #
# ================================================== #

from typing import Optional

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AUDIO,
    MODE_ASSISTANT,
    MODE_LLAMA_INDEX,
)
from pygpt_net.core.events import Event, AppEvent, KernelEvent, RenderEvent
from pygpt_net.core.bridge.context import BridgeContext, MultimodalContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Text:
    def __init__(self, window=None):
        """
        Text input controller

        :param window: Window instance
        """
        self.window = window
        self.ctx_pid = 0

    def send(
            self,
            text: str,
            reply: bool = False,
            internal: bool = False,
            prev_ctx: Optional[CtxItem] = None,
            parent_id: Optional[str] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ) -> CtxItem:
        """
        Send text message

        :param text: text to send
        :param reply: reply from plugins
        :param internal: internal call
        :param prev_ctx: previous context item (if reply)
        :param parent_id: parent context id
        :param multimodal_ctx: multimodal context
        :return: context item
        """
        self.window.update_status(trans('status.sending'))

        # event: prepare username
        event = Event(Event.USER_NAME, {
            'value': self.window.core.config.get('user_name'),
        })
        self.window.dispatch(event)
        user_name = event.data['value']

        # event: prepare ai.name
        event = Event(Event.AI_NAME, {
            'value': self.window.core.config.get('ai_name'),
        })
        self.window.dispatch(event)
        ai_name = event.data['value']

        # prepare mode, model, etc.
        mode = self.window.core.config.get('mode')
        model = self.window.core.config.get('model')
        model_data = self.window.core.models.get(model)
        stream_mode = self.window.core.config.get('stream')
        agent_provider = self.window.core.config.get('agent.llama.provider')
        agent_idx = self.window.core.config.get('agent.llama.idx')
        sys_prompt = self.window.core.config.get('prompt')
        sys_prompt_raw = sys_prompt  # store raw prompt (without addons)
        max_tokens = self.window.core.config.get('max_output_tokens')  # max output tokens
        base_mode = mode  # store parent mode
        functions = []  # functions to call
        tools_outputs = []  # tools outputs (assistant only)
        idx_mode = self.window.core.config.get('llama.idx.mode')

        # o1 models: disable stream mode
        if mode in [MODE_AGENT_LLAMA, MODE_AUDIO]:
            stream_mode = False
        if mode in [MODE_LLAMA_INDEX] and idx_mode == "retrieval":
            stream_mode = False

        # create ctx item
        ctx = CtxItem()
        ctx.meta = self.window.core.ctx.get_current_meta()  # CtxMeta (owner object)
        ctx.internal = internal
        ctx.current = True  # mark as current context item
        ctx.mode = mode  # store current selected mode (not inline changed)
        ctx.model = model  # store model list key, not real model id
        ctx.set_input(text, user_name)
        ctx.set_output(None, ai_name)
        ctx.prev_ctx = prev_ctx  # store previous context item if exists
        ctx.live = True
        ctx.pid = self.ctx_pid  # store PID
        self.ctx_pid += 1  # increment PID

        # if prev ctx is not empty, then copy input name to current ctx
        if prev_ctx is not None and prev_ctx.sub_call is True:  # sub_call = sent from expert
            ctx.input_name = prev_ctx.input_name

        # if reply from expert command
        if parent_id is not None:  # parent_id = reply from expert
            # At this point, ctx.meta ID = slave META ID (parent_id is given from slave, not from master)
            ctx.sub_reply = True  # mark as sub reply
            ctx.input_name = prev_ctx.input_name
            ctx.output_name = prev_ctx.output_name
        else:
            self.window.core.ctx.set_last_item(ctx)  # store last item

        self.window.controller.files.uploaded_ids = []  # clear uploaded files ids at the beginning

        # assistant: create thread, upload attachments
        if mode == MODE_ASSISTANT:
            self.window.controller.assistant.begin(ctx)

        # store in history (input only)
        if self.window.core.config.get('store_history'):
            self.window.core.history.append(ctx, "input")        

        self.window.controller.chat.log_ctx(ctx, "input")  # log

        # agent mode: before context
        if mode == MODE_AGENT:
            self.window.controller.agent.legacy.on_ctx_before(ctx)

        # event: context before
        event = Event(Event.CTX_BEFORE)
        event.ctx = ctx
        self.window.dispatch(event)

        # agent or expert mode
        sys_prompt = self.window.controller.agent.experts.append_prompts(mode, sys_prompt, parent_id)

        # on pre prompt event
        event = Event(Event.PRE_PROMPT, {
            'mode': mode,
            'value': str(sys_prompt),
        })
        self.window.dispatch(event)
        sys_prompt = event.data['value']

        # tool calls
        disable_native_func_calls = False
        if mode == MODE_LLAMA_INDEX:
            # check if index is selected
            if self.window.controller.idx.index_selected():
                disable_native_func_calls = True
                # ^^^ native func calls allowed only for LLM call, not for the query engine

        # build final prompt (+plugins)
        sys_prompt = self.window.core.prompt.prepare_sys_prompt(
            mode=mode,
            model=model_data,
            sys_prompt=sys_prompt,
            ctx=ctx,
            reply=reply,
            internal=internal,
            disable_native_tool_calls=disable_native_func_calls,
        )

        self.window.controller.chat.log("Appending input to chat window...")

        # render: begin
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
            "stream": stream_mode,
        }
        event = RenderEvent(RenderEvent.BEGIN, data)
        self.window.dispatch(event)

        # append text from input to chat window
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
        }
        event = RenderEvent(RenderEvent.INPUT_APPEND, data)
        self.window.dispatch(event)

        # add ctx to DB here and only update it after response,
        # MUST BE REMOVED AFTER AS FIRST MSG (LAST ON LIST)
        self.window.core.ctx.add(ctx, parent_id=parent_id)

        # update ctx list, but not reload all to prevent focus out on lists
        self.window.controller.ctx.update(
            reload=True,
            all=False,
        )

        # functions: prepare user and plugins functions (native mode only)
        functions += self.window.core.command.get_functions(parent_id)

        # assistant only
        if mode == MODE_ASSISTANT:
            # prepare tool outputs for assistant
            tools_outputs = self.window.controller.assistant.threads.handle_tool_outputs(ctx)
            if len(tools_outputs) > 0:
                self.window.controller.chat.log("Tool outputs sending...")

        # make API call
        try:
            # get attachments
            files = self.window.core.attachments.get_all(mode)
            num_files = len(files)
            if num_files > 0:
                self.window.controller.chat.log("Attachments ({}): {}".format(mode, num_files))

            # assistant
            assistant_id = self.window.core.config.get('assistant')
            if mode == MODE_AGENT_LLAMA:
                preset = self.window.controller.presets.get_current()
                if preset is not None:
                    assistant_id = preset.assistant_id

            self.window.dispatch(AppEvent(AppEvent.INPUT_CALL))  # app event
            bridge_context = BridgeContext(
                ctx=ctx,
                history=self.window.core.ctx.all(meta_id=parent_id),  # get all ctx items
                mode=mode,
                parent_mode=base_mode,
                model=model_data,
                system_prompt=sys_prompt,
                system_prompt_raw=sys_prompt_raw,  # for llama-index (query mode only)
                prompt=text,  # input text
                stream=stream_mode,  # is stream mode
                attachments=files,
                file_ids=self.window.controller.files.uploaded_ids,  # uploaded files IDs
                assistant_id=assistant_id,
                idx=self.window.controller.idx.current_idx,  # current idx
                idx_mode=idx_mode,  # llama index mode (chat or query)
                external_functions=functions,  # external functions
                tools_outputs=tools_outputs,  # if not empty then will submit outputs to assistant
                max_tokens=max_tokens,  # max output tokens
                multimodal_ctx=multimodal_ctx,  # multimodal context
            )
            extra = {
                'mode': mode,
                'reply': reply,
                'internal': internal,
            }
            if mode == MODE_AGENT_LLAMA:
                extra['agent_idx'] = agent_idx
                extra['agent_provider'] = agent_provider

            self.window.controller.chat.common.lock_input()  # lock input
            event = KernelEvent(KernelEvent.REQUEST, {
                'context': bridge_context,
                'extra': extra,
            })
            self.window.dispatch(event)

        except Exception as e:
            self.window.controller.chat.log("Bridge call ERROR: {}".format(e))  # log
            self.window.controller.chat.handle_error(e)
            print("Error when calling bridge: " + str(e))

        return ctx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List, Dict

from packaging.version import Version

from pygpt_net.item.assistant import AssistantFileItem, AssistantItem
from pygpt_net.provider.core.assistant_file.db_sqlite import DbSqliteProvider


class Files:
    def __init__(self, window=None):
        """
        Assistant files core

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.items = {}

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if data was patched
        """
        return self.provider.patch(app_version)

    def get(self, id: str) -> AssistantFileItem:
        """
        Get file item by file_id

        :param id: file ID
        :return: file item
        """
        if id in self.items:
            return self.items[id]

    def get_ids(self) -> List[str]:
        """
        Return all loaded file IDs

        :return: file ids list
        """
        return list(self.items.keys())


    def get_all(self) -> Dict[str, AssistantFileItem]:
        """
        Return all files

        :return: all files
        """
        return self.items

    def get_id_by_idx_all(self, idx: int) -> str:
        """
        Return file id by index

        :param idx: file idx
        :return: file id
        """
        return list(self.items.keys())[idx]


    def get_by_idx(self, idx: int) -> str:
        """
        Return file by index

        :param idx: file idx
        :return: model name
        """
        items = self.items
        return list(items.keys())[idx]


    def has(self, id: str) -> bool:
        """
        Check if file exists

        :param id: file id
        :return: True if file exists
        """
        return id in self.items

    def create(
            self,
            assistant: AssistantItem,
            thread_id: str,
            file_id: str,
            name: str,
            path: str,
            size: int) -> Optional[AssistantFileItem]:
        """
        Create new file

        :param assistant: assistant item
        :param thread_id: thread ID
        :param file_id: file ID
        :param name: file name
        :param path: file path
        :param size: file size
        :return: file item
        """
        file = AssistantFileItem()
        file.id = file_id
        file.file_id = file_id
        file.thread_id = thread_id
        file.name = name
        file.path = path
        file.size = size
        if assistant.vector_store is not None and assistant.vector_store != "":
            file.store_id = assistant.vector_store
        file.record_id = self.provider.create(file)
        self.items[file.id] = file
        return file

    def update(
            self,
            file: AssistantFileItem
    ) -> Optional[AssistantFileItem]:
        """
        Update file

        :param file: file
        :return: updated file or None if failed
        """
        self.items[file.id] = file
        self.provider.save(file)
        return file

    def get_names(self) -> Dict[str, str]:
        """
        Get file name mapping id <> name

        :return: file names
        """
        names = {}
        for id in self.items:
            file = self.items[id]
            names[id] = file.name
        return names

    def get_by_store_or_thread(
            self,
            store_id: str,
            thread_id: str
    ) -> Dict[str, AssistantFileItem]:
        """
        Get files by store or thread

        :param store_id: store ID
        :param thread_id: thread ID
        :return: files dict
        """
        return self.provider.get_by_store_or_thread(store_id, thread_id)

    def count_by_store_or_thread(
            self,
            store_id: str,
            thread_id: str
    ) -> int:
        """
        Count files by store or thread

        :param store_id: store ID
        :param thread_id: thread ID
        :return: files dict
        """
        return self.provider.count_by_store_or_thread(store_id, thread_id)

    def get_file_by_idx(
            self,
            idx: int,
            store_id: str,
            thread_id: str
    ) -> Optional[AssistantFileItem]:
        """
        Get file by list index

        :param idx: file index
        :param store_id: store ID
        :param thread_id: thread ID
        :return: file item
        """
        files = self.get_by_store_or_thread(store_id, thread_id)
        if idx >= len(files):
            return None
        return list(files.values())[idx]

    def get_file_id_by_idx(
            self,
            idx: int,
            store_id: str,
            thread_id: str
    ) -> Optional[str]:
        """
        Get file by list index

        :param idx: file index
        :param store_id: store ID
        :param thread_id: thread ID
        :return: file item
        """
        file = self.get_file_by_idx(idx, store_id, thread_id)
        if file is None:
            return None
        return file.file_id

    def get_all_by_file_id(self, file_id: str) -> dict:
        """
        Get all files by file ID

        :param file_id: file ID
        :return: files dict
        """
        return self.provider.get_all_by_file_id(file_id)

    def delete(self, file: AssistantFileItem) -> bool:
        """
        Delete file and remove from vector stores if exists

        :param file: file item
        :return: True if file was deleted
        """
        file_id = file.file_id
        items = self.get_all_by_file_id(file_id)  # get store_ids
        for id in items:
            store_id = items[id].store_id
            if store_id is None or store_id == "":
                continue  # skip if no store_id
            try:
                self.window.core.gpt.store.delete_store_file(store_id, file_id)  # remove from vector store
            except Exception as e:
                self.window.core.debug.log("Failed to delete file from vector store: " + str(e))
        self.provider.delete_by_id(file.record_id)  # delete file in DB
        try:
            self.window.core.gpt.store.delete_file(file.file_id)  # delete file in API
        except Exception as e:
            self.window.core.debug.log("Failed to delete remote file: " + str(e))
        if file.record_id in self.items:
            del self.items[file.record_id]
        return True

    def on_store_deleted(self, store_id: str):
        """
        Clear deleted store from files

        :param store_id: store ID
        """
        self.provider.clear_store_from_files(store_id)

    def on_all_stores_deleted(self):
        """Clear all deleted stores from files"""
        self.provider.clear_all_stores_from_files()

    def rename(self, record_id: int, name: str) -> bool:
        """
        Rename file

        :param record_id: record ID
        :param name: new name
        :return: True if renamed
        """
        self.provider.rename_file(record_id, name)
        if record_id in self.items:
            self.items[record_id].name = name
        return True

    def truncate(self, store_id: Optional[str] = None) -> bool:
        """
        Truncate all files

        :param store_id: store ID
        :return: True if truncated
        """
        if store_id is not None:
            self.window.core.gpt.store.remove_from_store(store_id)  # remove files from vector store
        else:
            self.window.core.gpt.store.remove_from_stores()  # remove files from all vector stores
        return self.truncate_local(store_id)  # truncate files in DB

    def truncate_local(self, store_id: Optional[str] = None) -> bool:
        """
        Truncate all files (local only)

        :param store_id: store ID
        :return: True if truncated
        """
        if store_id is not None:
            self.provider.truncate_by_store(store_id)  # truncate files in DB (by store_id)
        else:
            self.provider.truncate_all()  # truncate all files in DB
            self.items = {}  # clear items
        return True

    def import_from_store(self, store_id: str) -> bool:
        """
        Import files from store

        :param store_id: store ID
        :return: True if imported
        """
        files = self.window.core.gpt.store.import_store_files(store_id)
        for file in files:
            self.create(file.assistant, file.thread_id, file.file_id, file.name, file.path, file.size)
        return True

    def insert(self, store_id: str, data) -> AssistantFileItem:
        """
        Insert file object

        :param store_id: store ID
        :param data: file data from API
        """
        file = AssistantFileItem()
        file.id = data.id
        file.file_id = data.id
        file.thread_id = ""
        file.name = data.filename
        file.path = data.filename
        file.size = data.bytes
        file.store_id = store_id
        file.record_id = self.provider.create(file)
        self.items[file.id] = file
        return file

    def load(self):
        """Load file"""
        self.items = self.provider.load_all()
        self.sort_items()

    def sort_items(self):
        """Sort items"""
        return
        self.items = dict(sorted(self.items.items(), key=lambda item: item[0]))

    def save(self):
        """Save file"""
        self.provider.save_all(self.items)

    def get_version(self) -> str:
        """Get config version"""
        return self.provider.get_version()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        prompt = 'ADDITIONAL CONTEXT: I will provide you with additional context about my question. ' \
                 'When it is provided, then use this data as additional context and use it in your response. ' \
                 'Additional context will be prefixed with an "ADDITIONAL CONTEXT:" prefix. You can also provide a ' \
                 'command to query my context database anytime you need any additional context - to do this, return ' \
                 'to me the prepared prompt in JSON format, all in one line, using the following syntax: ' \
                 '~###~{"cmd": "get_context", "params": {"query": "<simple question here>"}}~###~. ' \
                 'Use ONLY this syntax and remember to surround the JSON string with ~###~. DO NOT use any other ' \
                 'syntax. When making query use language that I spoke to you.'

        plugin.add_option(
            "prompt",
            type="textarea",
            value=prompt,
            label="Prompt",
            description="Prompt used for instruct how to use additional data provided from Llama-index",
            tooltip="Prompt",
            advanced=True,
        )
        plugin.add_option(
            "idx",
            type="text",
            value="base",
            label="Indexes to use",
            description="ID's of indexes to use, default: base, separate by comma if you want to use "
                        "more than one index at once",
            tooltip="Index name",
        )
        plugin.add_option(
            "ask_llama_first",
            type="bool",
            value=False,
            label="Ask Llama-index first",
            description="When enabled, then Llama-index will be asked first, and response will be used "
                        "as additional knowledge in prompt. When disabled, then Llama-index will be "
                        "asked only when needed.",
        )
        plugin.add_option(
            "prepare_question",
            type="bool",
            value=False,
            label="Auto-prepare question before asking Llama-index first",
            description="When enabled, then question will be prepared before asking Llama-index first to create"
                        "best question for Llama-index.",
        )
        plugin.add_option(
            "model_prepare_question",
            type="combo",
            use="models",
            value="gpt-4o-mini",
            label="Model for question preparation",
            description="Model used to prepare question before asking Llama-index, default: gpt-3.5-turbo",
            tooltip="Model",
        )
        plugin.add_option(
            "prepare_question_max_tokens",
            type="int",
            value=500,
            label="Max output tokens for question preparation",
            description="Max tokens in output when preparing question before asking Llama-index",
            min=1,
            max=None,
        )
        plugin.add_option(
            "model_query",
            type="combo",
            value="gpt-4o-mini",
            label="Model",
            description="Model used for querying Llama-index, default: gpt-3.5-turbo",
            tooltip="Query model",
            use="models",
        )
        plugin.add_option(
            "max_question_chars",
            type="int",
            value=1000,
            label="Max characters in question",
            description="Max characters in question when querying Llama-index, 0 = no limit",
            min=0,
            max=None,
        )
        plugin.add_option(
            "append_meta",
            type="bool",
            value=False,
            label="Append metadata to context",
            description="If enabled, then metadata from Llama-index will be appended to additional context",
        )
        plugin.add_option(
            "syntax_prepare_question",
            type="textarea",
            value='Simplify the question into a short query for retrieving information from a vector store.',
            label="Prompt for question preparation",
            description="System prompt for question preparation",
            advanced=True,
        )
        plugin.add_cmd(
            "get_context",
            instruction="get additional context for a given query",
            params=[
                {
                    "name": "query",
                    "type": "str",
                    "description": "query to retrieve additional context for",
                    "required": True,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to get additional context for a given query",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import re

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenuBar, QVBoxLayout

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.ui.widget.dialog.base import BaseDialog
from pygpt_net.utils import trans

from .widgets import ToolWidget

class Tool:
    def __init__(self, window=None, tool=None):
        """
        HTML/JS canvas dialog

        :param window: Window instance
        :param tool: Tool instance
        """
        self.window = window
        self.tool = tool  # tool instance
        self.widget = ToolWidget(window, tool)
        self.layout = None
        self.menu_bar = None
        self.menu = {}
        self.actions = {}  # menu actions

    def setup_menu(self) -> QMenuBar:
        """
        Setup dialog menu

        :return: QMenuBar
        """
        # create menu bar
        self.menu_bar = QMenuBar()
        self.menu["file"] = self.menu_bar.addMenu(trans("menu.file"))

        self.actions["file.open"] = QAction(QIcon(":/icons/folder.svg"), trans("tool.html_canvas.menu.file.open"))
        self.actions["file.open"].triggered.connect(
            lambda: self.tool.open_file()
        )
        self.actions["file.save_as"] = QAction(QIcon(":/icons/save.svg"), trans("tool.html_canvas.menu.file.save_as"))
        self.actions["file.save_as"].triggered.connect(
            lambda: self.widget.output.signals.save_as.emit(
                re.sub(r'\n{2,}', '\n\n', self.widget.output.html_content), 'html')
        )
        self.actions["file.reload"] = QAction(QIcon(":/icons/reload.svg"), trans("tool.html_canvas.menu.file.reload"))
        self.actions["file.reload"].triggered.connect(
            lambda: self.tool.reload_output()
        )
        self.actions["file.clear"] = QAction(QIcon(":/icons/close.svg"), trans("tool.html_canvas.menu.file.clear"))
        self.actions["file.clear"].triggered.connect(
            lambda: self.tool.clear()
        )

        # add actions
        self.menu["file"].addAction(self.actions["file.open"])
        self.menu["file"].addAction(self.actions["file.save_as"])
        self.menu["file"].addAction(self.actions["file.reload"])
        self.menu["file"].addAction(self.actions["file.clear"])
        return self.menu_bar

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.widget.set_tab(tab)

    def setup(self):
        """Setup canvas dialog"""
        self.layout = self.widget.setup()
        self.layout.setMenuBar(self.setup_menu())  # add menu bar

        id = self.tool.get_dialog_id()
        dialog = ToolDialog(window=self.window, tool=self.tool)
        dialog.setLayout(self.layout)
        dialog.setWindowTitle(trans("dialog.html_canvas.title"))
        dialog.resize(800, 500)
        self.window.ui.dialog[id] = dialog

    def get_widget(self) -> ToolWidget:
        """
        Get widget

        :return: ToolWidget
        """
        return self.widget

    def get_tab(self) -> QVBoxLayout:
        """
        Get layout

        :return: QVBoxLayout
        """
        return self.layout


class ToolDialog(BaseDialog):
    def __init__(self, window=None, id="html_canvas", tool=None):
        """
        HTML canvas dialog

        :param window: main window
        :param id: logger id
        """
        super(ToolDialog, self).__init__(window, id)
        self.window = window
        self.tool = tool

    def closeEvent(self, event):
        """
        Close event

        :param event: close event
        """
        self.cleanup()
        super(ToolDialog, self).closeEvent(event)

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key press event
        """
        if event.key() == Qt.Key_Escape:
            self.cleanup()
            self.close()  # close dialog when the Esc key is pressed.
        else:
            super(ToolDialog, self).keyPressEvent(event)

    def cleanup(self):
        """Cleanup on close"""
        if self.window is None or self.tool is None:
            return
        self.tool.opened = False
        self.tool.close()
        self.tool.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.29 23:00:00                  #
# ================================================== #

from datetime import datetime

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.item.ctx import CtxItem
from pygpt_net.core.events import Event

from .config import Config


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "real_time"
        self.name = "Real Time"
        self.type = ["time"]
        self.description = "Appends current time and date to every system prompt."
        self.prefix = "Time"
        self.allowed_cmds = [
            "get_time",
        ]
        self.order = 2
        self.use_locale = True
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.POST_PROMPT_END:
            silent = False
            if 'silent' in data and data['silent']:
                silent = True
            data['value'] = self.on_system_prompt(
                data['value'],
                silent,
            )

        if name == Event.AGENT_PROMPT:
            silent = False
            if 'silent' in data and data['silent']:
                silent = True
            data['value'] = self.on_agent_prompt(
                data['value'],
                silent,
            )

        elif name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for option in self.allowed_cmds:
            if self.has_cmd(option):
                data['cmd'].append(self.get_cmd(option))  # append command

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        response = None
        for item in my_commands:
            try:
                if item["cmd"] in self.allowed_cmds and self.has_cmd(item["cmd"]):
                    # get time
                    if item["cmd"] == "get_time":
                        time = datetime.now().strftime('%A, %Y-%m-%d %H:%M:%S')
                        response = {
                            "request": {
                                "cmd": item["cmd"]
                            },
                            "result": time,
                        }
            except Exception as e:
                self.error(e)

        if response:
            self.reply(response, ctx)

    def on_system_prompt(self, prompt: str, silent: bool = False) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :param silent: silent mode (no logs)
        :return: updated prompt
        """
        if self.get_option_value("hour") or self.get_option_value("date"):
            if prompt.strip() != "":
                prompt += "\n\n"
            if self.get_option_value("hour") and self.get_option_value("date"):
                prompt += self.get_option_value("tpl").\
                    format(time=datetime.now().strftime('%A, %Y-%m-%d %H:%M:%S'))
            elif self.get_option_value("hour"):
                prompt += self.get_option_value("tpl").\
                    format(time=datetime.now().strftime('%H:%M:%S'))
            elif self.get_option_value("date"):
                prompt += self.get_option_value("tpl").\
                    format(time=datetime.now().strftime('%A, %Y-%m-%d'))
        return prompt

    def on_agent_prompt(self, prompt: str, silent: bool = False) -> str:
        """
        Event: AGENT_PROMPT

        :param prompt: prompt
        :param silent: silent mode (no logs)
        :return: updated prompt
        """
        if self.get_option_value("hour") or self.get_option_value("date"):
            if self.get_option_value("hour") and self.get_option_value("date"):
                prompt = self.get_option_value("tpl").\
                    format(time=datetime.now().strftime('%A, %Y-%m-%d %H:%M:%S')) + "\n\n" + prompt
            elif self.get_option_value("hour"):
                prompt = self.get_option_value("tpl").\
                    format(time=datetime.now().strftime('%H:%M:%S') + "\n\n" + prompt)
            elif self.get_option_value("date"):
                prompt = self.get_option_value("tpl").\
                    format(time=datetime.now().strftime('%A, %Y-%m-%d')) + "\n\n" + prompt
        return prompt





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.31 19:00:00                  #
# ================================================== #

from typing import Any

from pygpt_net.core.types import (
    MODE_AGENT,
)
from pygpt_net.core.events import KernelEvent, RenderEvent, Event
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.ctx.reply import ReplyContext
from pygpt_net.item.ctx import CtxItem


class Command:
    def __init__(self, window=None):
        """
        Command controller

        :param window: Window instance
        """
        self.window = window

    def handle(self, ctx: CtxItem):
        """
        Handle commands and expert mentions

        :param ctx: CtxItem
        """
        if self.window.controller.kernel.stopped():
            return

        mode = self.window.core.config.get('mode')

        # extract commands
        cmds = ctx.cmds_before  # from llama index tool calls pre-handler
        if not cmds:  # if no commands in context (from llama index tool calls)
            cmds = self.window.core.command.extract_cmds(ctx.output)

        if len(cmds) > 0:
            # check if commands are enabled, leave only enabled commands
            for cmd in cmds:
                cmd_id = str(cmd["cmd"])
                if not self.window.core.command.is_enabled(cmd_id):
                    self.log("Command not allowed: " + cmd_id)
                    cmds.remove(cmd)  # remove command from execution list
            if len(cmds) == 0:
                return  # abort if no commands

            ctx.cmds = cmds  # append commands to ctx
            self.log("Command call received...")

            # agent mode
            if mode == MODE_AGENT:
                commands = self.window.core.command.from_commands(cmds)  # pack to execution list
                self.window.controller.agent.legacy.on_cmd(
                    ctx,
                    commands,
                )

            # plugins
            self.log("Preparing command reply context...")

            reply = ReplyContext()
            reply.ctx = ctx
            reply.cmds = cmds
            if self.window.core.config.get('cmd'):
                reply.type = ReplyContext.CMD_EXECUTE
            else:
                reply.type = ReplyContext.CMD_EXECUTE_INLINE

            data = {
                "meta": ctx.meta,
            }
            event = RenderEvent(RenderEvent.TOOL_BEGIN, data)
            self.window.dispatch(event)  # show waiting
            context = BridgeContext()
            context.ctx = ctx
            context.reply_context = reply
            event = KernelEvent(KernelEvent.TOOL_CALL, {
                'context': context,
                'extra': {},
            })
            self.window.dispatch(event)

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.core.debug.info(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

import os

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem


class Analyzer:
    def __init__(self, window=None):
        """
        Image analyzer

        :param window: Window instance
        """
        self.window = window

    def send(
            self,
            ctx: CtxItem,
            prompt: str,
            files: dict
    ) -> str:
        """
        Send text from user input (called from UI)

        :param ctx: context
        :param prompt: analyze prompt
        :param files: files
        :return: response
        """
        model = self.window.core.models.get("gpt-4o")
        context = BridgeContext()
        context.prompt = prompt
        context.attachments = files
        context.history = []
        context.stream = False
        context.model = model
        context.system_prompt = ("You are an expert in image recognition. "
                                 "You are analyzing the image and providing a detailed description of the image.")

        extra = {}
        output = ""
        response = self.window.core.gpt.vision.send(context, extra)
        if response.choices[0] and response.choices[0].message.content:
            output = response.choices[0].message.content.strip()
        for id in files:
            ctx.images_before.append(files[id].path)
            files[id].consumed = True  # allow for deletion

        # re-allow clearing attachments
        self.window.controller.attachment.unlock()
        return output

    def from_screenshot(
            self,
            ctx: CtxItem,
            prompt: str
    ) -> str:
        """
        Image analysis from screenshot

        :param ctx: context
        :param prompt: analyze prompt
        :return: response
        """
        path = self.window.controller.painter.capture.screenshot(
            attach_cursor=True,
            silent=True,
        )
        attachment = AttachmentItem()
        attachment.path = path
        files = {
            "screenshot": attachment,
        }
        return self.send(ctx, prompt, files)

    def from_camera(
            self,
            ctx: CtxItem,
            prompt: str
    ) -> str:
        """
        Image analysis from camera

        :param ctx: context
        :param prompt: analyze prompt
        :return: response
        """
        path = self.window.controller.camera.capture_frame_save()
        attachment = AttachmentItem()
        attachment.path = path
        files = {
            "camera": attachment,
        }
        if path:
            return self.send(ctx, prompt, files)
        else:
            return "FAILED: There was a problem with capturing the image."

    def from_path(
            self,
            ctx: CtxItem,
            prompt: str,
            path: str
    ) -> str:
        """
        Image analysis from path

        :param ctx: context item
        :param prompt: analyze prompt
        :param path: path to file
        :return: response
        """
        if not path:
            return self.from_current_attachments(ctx, prompt)  # try current if no path provided

        if not os.path.exists(path):
            return "FAILED: File not found"

        attachment = AttachmentItem()
        attachment.path = path
        files = {
            "img": attachment,
        }
        return self.send(ctx, prompt, files)

    def from_current_attachments(
            self,
            ctx: CtxItem,
            prompt: str
    ) -> str:
        """
        Image analysis from current attachments

        :param ctx: context item
        :param prompt: analyze prompt
        :return: response
        """
        mode = self.window.core.config.get("mode")
        files = self.window.core.attachments.get_all(mode)  # clear is locked here
        result = self.send(ctx, prompt, files)  # unlocks clear

        # clear if capture clear
        if self.window.controller.attachment.is_capture_clear():
            self.window.controller.attachment.clear(True, auto=True)

        return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 03:00:00                  #
# ================================================== #

from pygpt_net.item.ctx import CtxItem

class Output:
    def __init__(self, plugin=None):
        self.plugin = plugin

    def handle(self, ctx: CtxItem, content) -> str:
        """
        Handle output

        :param ctx: CtxItem
        :param content: dict
        :return: HTML output
        """
        html = ""
        lang = "python"
        if isinstance(content, dict) and "code" in content:
            if "input" in content["code"]:
                if "lang" in content["code"]["input"] and "content" in content["code"]["input"]:
                    lang = content["code"]["input"]["lang"]
                    raw = "```{}\n{}\n```".format(lang, content["code"]["input"]["content"])
                    code = self.plugin.window.controller.chat.render.instance().parser.parse_code(raw)
                    html += "<div class=\"tool-output-header\">Input</div><div class=\"code-wrapper\">{}</div>".format(code)
            if "output" in content["code"] and "content" in content["code"]["output"]:
                if "lang" in content["code"]["output"]:
                    lang = content["code"]["output"]["lang"]
                    raw = "```{}\n{}\n```".format(lang, content["code"]["output"]["content"])
                    code = self.plugin.window.controller.chat.render.instance().parser.parse_code(raw)
                    html += "<div class=\"tool-output-header\">Output</div><div class=\"code-wrapper\">{}</div>".format(
                        code)
        return html


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.17 02:00:00                  #
# ================================================== #

import time

import numpy as np
import wave

from PySide6.QtMultimedia import QAudioFormat, QMediaDevices, QAudioSource
from PySide6.QtCore import QTimer

class Capture:
    def __init__(self, window=None):
        """
        Audio input capture core

        :param window: Window instance
        """
        self.window = window
        self.audio_source = None
        self.audio_io_device = None
        self.frames = []
        self.actual_audio_format = None
        self.path = None
        self.disconnected = False
        self.loop = False
        self.stop_callback = None
        self.start_time = 0
        self.devices = []
        self.selected_device = None
        self.bar = None
        self.check_audio_devices()

    def set_path(self, path: str):
        """
        Set audio input file path

        :param path: file path
        """
        self.path = path

    def set_bar(self, bar):
        """
        Set audio level bar

        :param bar: audio level bar
        """
        self.bar = bar

    def check_audio_devices(self):
        """Check audio input devices"""
        self.devices = QMediaDevices.audioInputs()
        if not self.devices:
            # no devices found
            self.selected_device = None
            print("No audio input devices found.")
        else:
            # set the first device as default
            self.selected_device = self.devices[0]

    def device_changed(self, index: int):
        """
        Change audio input device

        :param index: device index
        """
        if 0 <= index < len(self.devices):
            self.selected_device = self.devices[index]
        else:
            self.selected_device = None

    def prepare_device(self):
        """Set the current audio input device"""
        device_id = int(self.window.core.config.get('audio.input.device', 0))
        self.device_changed(device_id)

    def start(self):
        """
        Start audio input recording

        :return: True if started
        """
        # Clear previous frames
        self.frames = []

        # Prepare selected device
        self.prepare_device()
        if not self.selected_device:
            print("No audio input device selected")
            return
        if self.disconnected:
            print("Audio source disconnected, please connect the audio source")
            return False

        # Prevent multiple recordings
        if self.audio_source is not None:
            return False

        # Set up audio input and start recording
        self.setup_audio_input()
        self.start_time = time.time()
        return True

    def stop(self):
        """
        Stop audio input recording

        :return: True if stopped
        """
        result = False
        if self.audio_source is not None:
            # Disconnect the readyRead signal
            self.audio_io_device.readyRead.disconnect(self.process_audio_input)
            self.audio_source.stop()
            self.audio_source = None
            self.audio_io_device = None

            # Save frames to file (if any)
            if self.frames:
                self.save_audio_file(self.path)
                result = True
            else:
                print("No audio data recorded")
        return result

    def has_source(self) -> bool:
        """
        Check if audio source is available

        :return: True if available
        """
        if self.audio_source is not None:
            return True
        return False

    def has_frames(self) -> bool:
        """
        Check if audio frames are available

        :return: True if available
        """
        if self.frames:
            return True
        return False

    def get_frames(self) -> list:
        """
        Get recorded audio frames

        :return: list of frames
        """
        return self.frames

    def setup_audio_input(self):
        """Set up audio input device and start recording"""
        if not self.selected_device:
            print("No audio input device selected")
            return

        # Define audio format
        audio_format = QAudioFormat()
        audio_format.setSampleRate(int(self.window.core.config.get('audio.input.rate', 44100)))
        audio_format.setChannelCount(int(self.window.core.config.get('audio.input.channels', 1)))
        audio_format.setSampleFormat(QAudioFormat.SampleFormat.Int16)

        # Select default audio input device
        audio_input_device = self.selected_device

        # Check if the format is supported
        if not audio_input_device.isFormatSupported(audio_format):
            print("Requested format not supported, using nearest format.")
            audio_format = audio_input_device.preferredFormat()

        # Store the actual format being used
        self.actual_audio_format = audio_format

        # Create QAudioSource object with the device and format
        try:
            self.audio_source = QAudioSource(audio_input_device, audio_format)
        except Exception as e:
            self.disconnected = True
            print(f"Failed to create audio source: {e}")

        # Start audio input and obtain the QIODevice
        try:
            self.audio_io_device = self.audio_source.start()
            if self.audio_io_device is None:
                raise Exception("Unable to access audio input device")
        except Exception as e:
            print(f"Failed to start audio input: {e}")
            self.disconnected = True
            self.audio_source = None
            self.audio_io_device = None
            return

        # Connect the readyRead signal to process incoming data
        self.audio_io_device.readyRead.connect(self.process_audio_input)

    def process_audio_input(self):
        """Process incoming audio data"""
        # add seconds to stop timer
        data = self.audio_io_device.readAll()
        if data.isEmpty():
            return

        # Convert QByteArray to bytes
        data_bytes = data.data()

        # Append raw data to frames list for saving
        self.frames.append(data_bytes)

        # Determine the correct dtype and normalization factor
        sample_format = self.actual_audio_format.sampleFormat()
        dtype = self.get_dtype_from_sample_format(sample_format)
        normalization_factor = self.get_normalization_factor(sample_format)

        # Convert bytes to NumPy array of the appropriate type
        samples = np.frombuffer(data_bytes, dtype=dtype)
        if samples.size == 0:
            return

        # For unsigned formats, center the data
        if sample_format == QAudioFormat.SampleFormat.UInt8:
            samples = samples.astype(np.int16)
            samples -= 128

        # Compute RMS of the audio samples as float64 for precision
        rms = np.sqrt(np.mean(samples.astype(np.float64) ** 2))

        # Normalize RMS value based on the sample format
        level = rms / normalization_factor

        # Ensure level is within 0.0 to 1.0
        level = min(max(level, 0.0), 1.0)

        # Scale to 0-100
        level_percent = level * 100

        # Update the level bar widget
        self.update_audio_level(level_percent)

        # Handle loop recording
        if self.loop and self.stop_callback is not None:
            stop_interval = int(self.window.core.config.get('audio.input.stop_interval', 10))
            current_time = time.time()
            time_elapsed = current_time - self.start_time
            if time_elapsed >= stop_interval:
                # stop recording here, save audio chunk to WAV file, run transcription, and start recording again
                self.start_time = current_time
                QTimer.singleShot(0, self.stop_callback)  # required QTimer to prevent crash!!!

    def update_audio_level(self, level: int):
        """
        Update the audio level bar

        :param level: audio level
        """
        if self.bar is not None:
            self.bar.setLevel(level)

    def reset_audio_level(self):
        """Reset the audio level bar"""
        if self.bar is not None:
            self.bar.setLevel(0)

    def save_audio_file(self, filename: str):
        """
        Save the recorded audio frames to a WAV file

        :param filename: output file name
        """
        # Define the parameters for the WAV file
        channels = self.actual_audio_format.channelCount()
        sample_size = self.actual_audio_format.bytesPerSample()
        frame_rate = self.actual_audio_format.sampleRate()

        # Open the WAV file
        wf = wave.open(filename, 'wb')
        wf.setnchannels(channels)
        wf.setsampwidth(sample_size)
        wf.setframerate(frame_rate)

        # Write frames to the file
        wf.writeframes(b''.join(self.frames))
        wf.close()

    def get_dtype_from_sample_format(self, sample_format):
        """
        Get the NumPy dtype corresponding to the QAudioFormat sample format

        :param sample_format: QAudioFormat.SampleFormat
        """
        if sample_format == QAudioFormat.SampleFormat.UInt8:
            return np.uint8
        elif sample_format == QAudioFormat.SampleFormat.Int16:
            return np.int16
        elif sample_format == QAudioFormat.SampleFormat.Int32:
            return np.int32
        elif sample_format == QAudioFormat.SampleFormat.Float:
            return np.float32
        else:
            raise ValueError("Unsupported sample format")

    def get_normalization_factor(self, sample_format):
        """
        Get the normalization factor for the QAudioFormat sample format

        :param sample_format: QAudioFormat.SampleFormat
        """
        if sample_format == QAudioFormat.SampleFormat.UInt8:
            return 255.0
        elif sample_format == QAudioFormat.SampleFormat.Int16:
            return 32768.0
        elif sample_format == QAudioFormat.SampleFormat.Int32:
            return float(2 ** 31)
        elif sample_format == QAudioFormat.SampleFormat.Float:
            return 1.0
        else:
            raise ValueError("Unsupported sample format")

    def check_audio_input(self) -> bool:
        """
        Check if default audio input device is working

        :return: True if working
        """
        import pyaudio
        p = pyaudio.PyAudio()
        try:
            rate = 44100
            channels = 1
            stream = p.open(format=pyaudio.paInt16,
                            channels=channels,
                            rate=rate,
                            input=True)
            stream.stop_stream()
            stream.close()
            p.terminate()
            return True
        except Exception as e:
            p.terminate()
            return False


# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_IMAGE,
    MODE_VISION,
)

class Vision:
    def __init__(self, window=None):
        """
        UI vision update controller

        :param window: Window instance
        """
        self.window = window

    def has_vision(self):
        """
        Check if vision is available

        :return: True if vision is available
        """
        mode = self.window.core.config.get("mode")
        if mode == MODE_IMAGE:
            return False
        if mode == MODE_VISION:
            return True
        if self.window.controller.plugins.is_type_enabled('vision'):
            return True
        if self.is_vision_model() and mode in [MODE_CHAT]:
            return True
        return False

    def update(self):
        """Update vision options"""
        if self.window.controller.painter.is_active():
            self.window.controller.camera.setup()
            self.window.controller.camera.show_camera()
        else:
            if self.has_vision():
                self.window.controller.camera.setup()
                self.window.controller.camera.show_camera()
                self.window.controller.chat.vision.show_inline()
            else:
                self.window.controller.camera.hide_camera()
                self.window.controller.chat.vision.hide_inline()

    def is_vision_model(self) -> bool:
        """
        Check if current model has vision capabilities

        :return: True if vision model
        """
        model_id = self.window.core.config.get("model")
        if model_id is not None:
            if self.window.core.models.has(model_id):
                model = self.window.core.models.get(model_id)
                if MODE_VISION in model.mode:
                    return True
        return False




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 17:00:00                  #
# ================================================== #

from datetime import datetime
from typing import Dict, Any


class Tab:

    # types
    TAB_ADD = -1
    TAB_CHAT = 0
    TAB_NOTEPAD = 1
    TAB_FILES = 2
    TAB_TOOL_PAINTER = 3
    TAB_TOOL_CALENDAR = 4
    TAB_TOOL = 100

    def __init__(self):
        """
        Tab data
        """
        self.uuid = None
        self.pid = None
        self.idx = 0
        self.type = 0
        self.title = ""
        self.icon = None
        self.tooltip = None
        self.data_id = None
        self.new_idx = None
        self.custom_name = False
        self.child = None
        self.parent = None
        self.column_idx = 0
        self.tool_id = None

        dt = datetime.now()
        self.created_at = dt
        self.updated_at = dt

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to dict

        :return: dict
        """
        return {
            "uuid": str(self.uuid),
            "pid": self.pid,
            "idx": self.idx,
            "type": self.type,
            "title": self.title,
            "icon": self.icon,
            "tooltip": self.tooltip,
            "data_id": self.data_id,
            "child": str(self.child),  # child widget
            "parent": str(self.parent),  # parent column
            "custom_name": self.custom_name,
            "custom_idx": self.new_idx,
            "created_at": str(self.created_at),
            "updated_at": str(self.updated_at),
            "column_idx": self.column_idx,
            "tool_id": self.tool_id,
        }

    def __str__(self) -> str:
        """
        String representation

        :return: str
        """
        return str(self.to_dict())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import json
from typing import Optional, Dict, Any

from pygpt_net.item.ctx import CtxItem
from .base import BaseEvent

class Event(BaseEvent):

    # Events
    AI_NAME = "ai.name"
    AGENT_PROMPT = "agent.prompt"  # on agent prompt
    AUDIO_INPUT_RECORD_START = "audio.input.record.start"  # start audio input recording
    AUDIO_INPUT_RECORD_STOP = "audio.input.record.stop"  # stop audio input recording
    AUDIO_INPUT_RECORD_TOGGLE = "audio.input.record.toggle"  # toggle audio input recording
    AUDIO_INPUT_TRANSCRIBE = "audio.input.transcribe"  # transcribe audio file
    AUDIO_INPUT_STOP = "audio.input.stop"
    AUDIO_INPUT_TOGGLE = "audio.input.toggle"
    AUDIO_OUTPUT_STOP = "audio.output.stop"
    AUDIO_OUTPUT_TOGGLE = "audio.output.toggle"
    AUDIO_PLAYBACK = "audio.playback"
    AUDIO_READ_TEXT = "audio.read_text"
    CMD_EXECUTE = "cmd.execute"
    CMD_INLINE = "cmd.inline"
    CMD_SYNTAX = "cmd.syntax"
    CMD_SYNTAX_INLINE = "cmd.syntax.inline"
    CTX_AFTER = "ctx.after"
    CTX_BEFORE = "ctx.before"
    CTX_BEGIN = "ctx.begin"
    CTX_END = "ctx.end"
    CTX_SELECT = "ctx.select"
    DISABLE = "disable"
    ENABLE = "enable"
    FORCE_STOP = "force.stop"
    INPUT_BEFORE = "input.before"
    MODE_BEFORE = "mode.before"
    MODE_SELECT = "mode.select"
    MODEL_BEFORE = "model.before"
    MODEL_SELECT = "model.select"
    PLUGIN_SETTINGS_CHANGED = "plugin.settings.changed"
    PLUGIN_OPTION_GET = "plugin.option.get"
    POST_PROMPT = "post.prompt"
    POST_PROMPT_ASYNC = "post.prompt.async"
    POST_PROMPT_END = "post.prompt.end"
    PRE_PROMPT = "pre.prompt"
    SYSTEM_PROMPT = "system.prompt"
    TOOL_OUTPUT_RENDER = "tool.output.render"
    UI_ATTACHMENTS = "ui.attachments"
    UI_VISION = "ui.vision"
    USER_NAME = "user.name"
    USER_SEND = "user.send"

    def __init__(
            self,
            name: Optional[str] = None,
            data: Optional[dict] = None,
            ctx: Optional[CtxItem] = None
    ):
        """
        Event object class

        :param name: event name
        :param data: event data
        :param ctx: context instance
        """
        super(Event, self).__init__(name, data, ctx)
        self.id = "Event"
        self.name = name
        self.data = data
        if self.data is None:
            self.data = {}
        self.ctx = ctx  # CtxItem
        self.stop = False  # True to stop propagation
        self.internal = False
        # internal event, not called from user
        # internal event is handled synchronously, ctx item has internal flag

    def to_dict(self) -> Dict[str, Any]:
        """
        Dump event to dict

        :return: Event dict
        """
        return {
            'name': self.name,
            'data': self.data,
            'ctx': self.ctx.to_dict() if self.ctx else None,
            'stop': self.stop,
            'internal': self.internal,
        }

    def dump(self) -> str:
        """
        Dump event to json string

        :return: JSON string
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self) -> str:
        """
        String representation of event

        :return: Event string
        """
        return self.dump()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any

from llama_index.agent.openai import OpenAIAgent as Agent

from .base import BaseAgent

class OpenAIAgent(BaseAgent):
    def __init__(self, *args, **kwargs):
        super(OpenAIAgent, self).__init__(*args, **kwargs)
        self.id = "openai"
        self.mode = "step"  # step|plan

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        tools = kwargs.get("tools", [])
        verbose = kwargs.get("verbose", False)
        llm = kwargs.get("llm", None)
        chat_history = kwargs.get("chat_history", [])
        max_iterations = kwargs.get("max_iterations", 10)
        return Agent.from_tools(
            tools=tools,
            llm=llm,
            chat_history=chat_history,
            max_iterations=max_iterations,
            verbose=verbose,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from typing import Any

from PySide6.QtWidgets import QApplication

from pygpt_net.core.events import RenderEvent
from pygpt_net.item.ctx import CtxItem


class Stream:
    def __init__(self, window=None):
        """
        Stream controller

        :param window: Window instance
        """
        self.window = window

    def append(self, ctx: CtxItem):
        """
        Handle stream response

        :param ctx: CtxItem
        """
        output = ""
        output_tokens = 0
        begin = True
        error = None

        # chunks: stream begin
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
        }
        event = RenderEvent(RenderEvent.STREAM_BEGIN, data)
        self.window.dispatch(event)

        # read stream
        try:
            if ctx.stream is not None:
                self.log("[chat] Stream begin...")  # log

                tool_calls = []

                for chunk in ctx.stream:
                    # if force stop then break
                    if self.window.controller.kernel.stopped():
                        break

                    if error is not None:
                        break  # break if error

                    response = None
                    chunk_type = "raw"
                    if (hasattr(chunk, 'choices')
                            and chunk.choices[0] is not None
                            and hasattr(chunk.choices[0], 'delta')
                            and chunk.choices[0].delta is not None):
                        chunk_type = "api_chat"
                    elif (hasattr(chunk, 'choices')
                          and chunk.choices[0] is not None
                          and hasattr(chunk.choices[0], 'text')
                          and chunk.choices[0].text is not None):
                        chunk_type = "api_completion"
                    elif (hasattr(chunk, 'content')
                          and chunk.content is not None):
                        chunk_type = "langchain_chat"
                    elif (hasattr(chunk, 'delta')
                          and chunk.delta is not None):
                        chunk_type = "llama_chat"

                    # OpenAI chat completion
                    if chunk_type == "api_chat":
                        citations = None
                        if chunk.choices[0].delta and chunk.choices[0].delta.content is not None:
                            if citations is None:
                                if chunk and hasattr(chunk, 'citations') and chunk.citations is not None:
                                    citations = chunk.citations
                                    ctx.urls = citations
                            response = chunk.choices[0].delta.content
                        if chunk.choices[0].delta and chunk.choices[0].delta.tool_calls:
                            tool_chunks = chunk.choices[0].delta.tool_calls
                            for tool_chunk in tool_chunks:
                                if len(tool_calls) <= tool_chunk.index:
                                    tool_calls.append(
                                        {
                                            "id": "",
                                            "type": "function",
                                            "function": {
                                                "name": "",
                                                "arguments": ""
                                            }
                                        }
                                    )
                                tool_call = tool_calls[tool_chunk.index]
                                if tool_chunk.id:
                                    tool_call["id"] += tool_chunk.id
                                if tool_chunk.function.name:
                                    tool_call["function"]["name"] += tool_chunk.function.name
                                if tool_chunk.function.arguments:
                                    tool_call["function"]["arguments"] += tool_chunk.function.arguments

                    # OpenAI completion
                    elif chunk_type == "api_completion":
                        if chunk.choices[0].text is not None:
                            response = chunk.choices[0].text

                    # langchain chat
                    elif chunk_type == "langchain_chat":
                        if chunk.content is not None:
                            response = str(chunk.content)

                    # llama chat
                    elif chunk_type == "llama_chat":
                        if chunk.delta is not None:
                            response = str(chunk.delta)
                        tool_chunks = chunk.message.additional_kwargs.get("tool_calls", [])
                        if tool_chunks:
                            for tool_chunk in tool_chunks:
                                args = tool_chunk.function.arguments
                                if not args:
                                    args = "{}"  # JSON encoded
                                tool_call = {
                                        "id": tool_chunk.id,
                                        "type": "function",
                                        "function": {
                                            "name": tool_chunk.function.name,
                                            "arguments": args
                                        }
                                    }
                                tool_calls.clear()
                                tool_calls.append(tool_call)

                    # raw text: llama-index and langchain completion
                    else:
                        if chunk is not None:
                            response = str(chunk)

                    if response is not None and response != "":
                        if begin and response == "":  # prevent empty beginning
                            continue
                        output += response
                        output_tokens += 1
                        data = {
                            "meta": ctx.meta,
                            "ctx": ctx,
                            "chunk": response,
                            "begin": begin,
                        }
                        event = RenderEvent(RenderEvent.STREAM_APPEND, data)
                        self.window.dispatch(event)
                        QApplication.processEvents()  # process events to update UI after each chunk
                        begin = False

                # unpack and store tool calls
                if tool_calls:
                    self.window.core.command.unpack_tool_calls_chunks(ctx, tool_calls)

        except Exception as e:
            self.window.core.debug.log(e)
            error = e

        self.window.controller.ui.update_tokens()  # update UI tokens

        # update ctx
        ctx.output = output
        ctx.set_tokens(ctx.input_tokens, output_tokens)

        # chunks: stream end
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
        }
        event = RenderEvent(RenderEvent.STREAM_END, data)
        self.window.dispatch(event)

        # log
        self.log("[chat] Stream end.")

        if error is not None:
            raise error  # raise error if any, to display in UI

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.core.debug.info(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional

from .base import BaseEvent


class KernelEvent(BaseEvent):
    """Kernel events"""

    # core, events sent from kernel
    INIT = "kernel.init"
    RESTART = "kernel.restart"
    STOP = "kernel.stop"
    TERMINATE = "kernel.terminate"

    # input
    INPUT_SYSTEM = "kernel.input.system"
    INPUT_USER = "kernel.input.user"

    # queue
    AGENT_CALL = "kernel.agent.call"
    AGENT_CONTINUE = "kernel.agent.continue"
    APPEND_BEGIN = "kernel.append.begin"
    APPEND_DATA = "kernel.append.data"
    APPEND_END = "kernel.append.end"
    CALL = "kernel.call"
    REQUEST = "kernel.request"
    REQUEST_NEXT = "kernel.request.next"
    RESPONSE_ERROR = "kernel.response.error"
    RESPONSE_FAILED = "kernel.response.failed"
    RESPONSE_OK = "kernel.response.OK"
    TOOL_CALL = "kernel.tool.call"

    # reply
    REPLY_ADD = "kernel.reply.add"
    REPLY_RETURN = "kernel.reply.return"

    # state
    STATE = "kernel.state"
    STATE_BUSY = "kernel.state.busy"
    STATE_IDLE = "kernel.state.idle"
    STATE_ERROR = "kernel.state.error"
    STATUS = "kernel.status"

    def __init__(
            self,
            name: Optional[str] = None,
            data: Optional[dict] = None,
    ):
        """
        Event object class

        :param name: event name
        :param data: event data
        """
        super(KernelEvent, self).__init__(name, data)
        self.id = "KernelEvent"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.01 11:00:00                  #
# ================================================== #

from typing import Optional, List

from PySide6.QtCore import Slot

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.types import MODE_IMAGE
from pygpt_net.item.ctx import CtxItem
from pygpt_net.core.events import Event, KernelEvent, RenderEvent
from pygpt_net.utils import trans


class Image:
    def __init__(self, window=None):
        """
        Image controller

        :param window: Window instance
        """
        self.window = window

    def send(
            self,
            text: str,
            prev_ctx: Optional[CtxItem] = None,
            parent_id: Optional[int] = None,
    ) -> CtxItem:
        """
        Send prompt for image generate

        :param text: prompt for image generation
        :param prev_ctx: previous ctx item
        :param parent_id: parent ctx id
        :return: ctx item
        """
        num = int(self.window.ui.config['global']['img_variants'].input.text() or 1)
        if num < 1:
            num = 1
        elif num > 4:
            num = 4

        # force 1 image if dall-e-3 model is used
        model = self.window.core.config.get('model')
        model_data = self.window.core.models.get(model)
        if model_data.id == 'dall-e-3':
            num = 1

        self.window.update_status(trans('status.sending'))

        # create ctx item
        ctx = CtxItem()
        ctx.current = True  # mark as current context item
        ctx.meta = self.window.core.ctx.get_current_meta()  # CtxMeta (owner object)
        ctx.set_input(text, self.window.core.config.get('user_name'))
        ctx.prev_ctx = prev_ctx  # store previous context item
        ctx.live = True

        # event: context before
        event = Event(Event.CTX_BEFORE)
        event.ctx = ctx
        self.window.dispatch(event)

        # add ctx to DB
        self.window.core.ctx.add(ctx)

        # render: begin
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
            "stream": False,
        }
        event = RenderEvent(RenderEvent.STREAM_BEGIN, data)
        self.window.dispatch(event)

        # append input to chat
        data = {
            "meta": ctx.meta,
            "ctx": ctx,
        }
        event = RenderEvent(RenderEvent.INPUT_APPEND, data)
        self.window.dispatch(event)

        # handle ctx name (generate title from summary if not initialized)
        if self.window.core.config.get('ctx.auto_summary'):
            self.window.controller.ctx.prepare_name(ctx)

        # generate image
        bridge_context = BridgeContext(
            ctx=ctx,
            mode=MODE_IMAGE,
            model=model_data,  # model instance
            prompt=text,
        )
        extra = {
            "num": num,
        }
        try:
            event = KernelEvent(KernelEvent.REQUEST, {
                'context': bridge_context,
                'extra': extra,
            })
            self.window.dispatch(event)
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)
            self.window.update_status(trans('status.error'))

        return ctx

    @Slot(object, list, str)
    def handle_response(
            self,
            ctx: CtxItem,
            paths: List[str],
            prompt: str
    ):
        """
        Handle response

        :param ctx: ctx item
        :param paths: list with paths to downloaded images
        :param prompt: prompt used to generate images
        """
        self.window.dispatch(KernelEvent(KernelEvent.STATE_IDLE, {
            "id": "img",
        }))
        string = ""
        i = 1
        for path in paths:
            string += "{}) `{}`".format(i, path) + "\n"
            i += 1

        if self.window.core.config.get('img_dialog_open'):
            self.window.tools.get("viewer").open_images(paths)  # use viewer tool

        if not self.window.core.config.get('img_raw'):
            string += "\nPrompt: "
            string += prompt

        local_urls = self.window.core.filesystem.make_local_list(paths)
        ctx.images = local_urls  # save images paths
        ctx.set_output(string.strip())

        # event: after context
        event = Event(Event.CTX_AFTER)
        event.ctx = ctx
        self.window.dispatch(event)

        # store last mode (in text mode this is handled in send_text)
        mode = self.window.core.config.get('mode')
        self.window.core.ctx.post_update(mode)  # post update context, store last mode, etc.
        self.window.core.ctx.store()  # save current ctx to DB
        self.window.update_status(trans('status.img.generated'))

        # update ctx in DB
        self.window.core.ctx.update_item(ctx)

        self.window.controller.chat.common.unlock_input()  # unlock input

    @Slot(object, list, str)
    def handle_response_inline(
            self,
            ctx: CtxItem,
            paths: List[str],
            prompt: str
    ):
        """
        Handle inline response

        :param ctx: ctx item
        :param paths: list with paths to downloaded images
        :param prompt: prompt used to generate images
        """
        self.window.dispatch(KernelEvent(KernelEvent.STATE_IDLE, {
            "id": "img",
        }))
        string = ""
        i = 1
        for path in paths:
            string += "{}) `{}`".format(i, path) + "\n"
            i += 1

        local_urls = self.window.core.filesystem.make_local_list(paths)
        ctx.images = local_urls  # save images paths in ctx item here
        
        self.window.core.ctx.update_item(ctx)  # update in DB
        self.window.update_status(trans('status.img.generated'))  # update status

        # WARNING:
        # if internal (sync) mode, then re-send OK status response, if not, append only img result
        # it will only inform system that image was generated, user will see it in chat with image after render
        # of ctx item (link to images are appended to ctx item)
        if ctx.internal:
            ctx.results.append(
                {
                    "request": {
                        "cmd": "image",
                    },
                    "result": "OK. Generated {} image(s).".format(len(paths)),
                    "paths": paths,
                }
            )
            ctx.reply = False
            data = {
                "meta": ctx.meta,
                "ctx": ctx,
            }
            event = RenderEvent(RenderEvent.EXTRA_APPEND, data)
            self.window.dispatch(event)  # show image first

            event = RenderEvent(RenderEvent.EXTRA_END, data)
            self.window.dispatch(event)  # end extra

            context = BridgeContext()
            context.ctx = ctx
            extra = {
                "flush": True,
            }
            event = KernelEvent(KernelEvent.REPLY_ADD, {
                'context': context,
                'extra': extra,
            })
            self.window.dispatch(event)
            self.window.controller.chat.common.unlock_input()  # unlock input

            event = RenderEvent(RenderEvent.TOOL_UPDATE, data)
            self.window.dispatch(event)  # end of tool, hide spinner icon
            return

        # NOT internal-mode, user called, so append only img output to chat (show images now):

        data = {
            "meta": ctx.meta,
            "ctx": ctx,
        }
        event = RenderEvent(RenderEvent.EXTRA_APPEND, data)
        self.window.dispatch(event)  # show image first

        event = RenderEvent(RenderEvent.EXTRA_END, data)
        self.window.dispatch(event)  # end extra

        self.window.controller.chat.common.unlock_input()  # unlock input

        event = RenderEvent(RenderEvent.TOOL_UPDATE, data)
        self.window.dispatch(event)  # end of tool, hide spinner icon



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.25 02:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal, QObject

from pygpt_net.core.events import RenderEvent
from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals
from pygpt_net.utils import trans

class Builder(QObject):
    def __init__(self, plugin=None):
        super(Builder, self).__init__()
        self.plugin = plugin
        self.worker = None

    def build_and_restart(self):
        """Run IPython image build and restart container"""
        self.build_image(restart=True)

    def build_image(self, restart: bool = False):
        """
        Run IPython image build

        :param restart: Restart container
        """
        try:
            self.worker = Worker()
            self.worker.plugin = self.plugin
            self.worker.restart = restart
            self.worker.signals.build_finished.connect(self.handle_build_finished)
            self.worker.signals.error.connect(self.handle_build_failed)
            self.plugin.window.threadpool.start(self.worker)
        except Exception as e:
            self.plugin.window.ui.dialogs.alert(e)

    @Slot()
    def handle_build_finished(self):
        """Handle build finished"""
        self.plugin.window.ui.dialogs.alert(trans('ipython.docker.build.finish'))
        self.plugin.window.update_status(trans('ipython.docker.build.finish'))
        self.plugin.window.controller.kernel.stop()
        event = RenderEvent(RenderEvent.END)
        self.plugin.window.dispatch(event)

    @Slot(object)
    def handle_build_failed(self, error):
        """Handle build failed"""
        self.plugin.window.ui.dialogs.alert(str(error))
        self.plugin.window.update_status(str(error))
        self.plugin.window.controller.kernel.stop()
        event = RenderEvent(RenderEvent.END)
        self.plugin.window.dispatch(event)

class WorkerSignals(BaseSignals):
    build_finished = Signal()

class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.restart = False

    @Slot()
    def run(self):
        try:
            self.plugin.get_interpreter().build_image()
            if self.restart:
                self.plugin.get_interpreter().restart()
            self.signals.build_finished.emit()
        except Exception as e:
            self.signals.error.emit(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.23 21:00:00                  #
# ================================================== #

import copy
import datetime
import os
import re

from pathlib import Path
from packaging.version import Version

from pygpt_net.core.profile import Profile
from pygpt_net.provider.core.config.json_file import JsonFileProvider


class Config:
    CONFIG_DIR = 'pygpt-net'
    SNAP_NAME = 'pygpt'
    TYPE_STR = 0
    TYPE_INT = 1
    TYPE_FLOAT = 2
    TYPE_BOOL = 3

    def __init__(self, window=None):
        """
        Config handler

        :param window: Window instance
        """
        version = self.get_version()
        self.window = window
        self.profile = Profile(window)
        self.profile.version = version
        self.profile.init(Config.get_base_workdir())
        self.path = None
        self.initialized = False
        self.initialized_base = False
        self.initialized_workdir = False
        self.db_echo = False
        self.data = {}  # user config
        self.data_base = {}  # base config
        self.data_session = {}  # temporary config (session only)
        self.version = version
        self.dirs = {
            "capture": "capture",
            "css": "css",
            "data": "data",
            "history": "history",
            "idx": "idx",
            "ctx_idx": "ctx_idx",
            "img": "img",
            "locale": "locale",
            "presets": "presets",
            "upload": "upload",
            "tmp": "tmp",
        }
        self.provider = JsonFileProvider(window)
        self.provider.path_app = self.get_app_path()
        self.provider.meta = self.append_meta()

        if not self.initialized_workdir:
            workdir = self.prepare_workdir()
            self.set_workdir(workdir)

    def is_compiled(self) -> bool:
        """
        Return True if compiled version

        :return: True if compiled version
        """
        return __file__.endswith('.pyc')

    def install(self):
        """Install database and provider data"""
        self.window.core.db.echo = self.db_echo  # verbose on/off
        self.window.core.db.init()

        # install provider configs
        self.provider.install()

    def get_path(self) -> str:
        """
        Return workdir path

        :return: workdir path
        """
        if not self.initialized_workdir:
            workdir = self.prepare_workdir()
            self.set_workdir(workdir)
        return self.path

    @staticmethod
    def get_base_workdir() -> str:
        """
        Return base workdir path

        :return: base workdir path
        """
        path = os.path.join(Path.home(), '.config', Config.CONFIG_DIR)
        if "PYGPT_WORKDIR" in os.environ and os.environ["PYGPT_WORKDIR"] != "":
            print("FORCE using workdir: {}".format(os.environ["PYGPT_WORKDIR"]))
            # convert relative path to absolute path if needed
            if not os.path.isabs(os.environ["PYGPT_WORKDIR"]):
                path = os.path.join(os.getcwd(), os.environ["PYGPT_WORKDIR"])
            else:
                path = os.environ["PYGPT_WORKDIR"]
            if not os.path.exists(path):
                print("Workdir path not exists: {}".format(path))
                print("Creating workdir path: {}".format(path))
                os.makedirs(path, exist_ok=True)
        return path

    @staticmethod
    def prepare_workdir() -> str:
        """
        Prepare workdir

        :return: workdir path
        """
        is_test = os.environ.get('ENV_TEST') == '1'
        path = Path(Config.get_base_workdir())
        if not path.exists() and not is_test:  # DISABLE in tests!!!
            path.mkdir(parents=True, exist_ok=True)
        path_file = "path.cfg"
        p = os.path.join(str(path), path_file)
        if not os.path.exists(p) and not is_test:  # DISABLE in tests!!!
            with open(p, 'w', encoding='utf-8') as f:
                f.write("")
        else:
            with open(p, 'r', encoding='utf-8') as f:
                tmp_path = f.read().strip()
            if tmp_path:
                tmp_path = tmp_path.replace("%HOME%", str(Path.home()))
                if os.path.exists(tmp_path):
                    path = Path(tmp_path)
                else:
                    print("CRITICAL: Workdir path not exists: {}".format(tmp_path))
        return str(path)

    def set_workdir(self, path: str, reload: bool = False):
        """
        Update workdir path

        :param path: new path
        :param reload: reload config
        """
        self.path = path
        self.provider.path = path
        if reload:
            self.initialized = False
            self.init(True)

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if patched
        """
        return self.provider.patch(app_version)

    def get_user_dir(self, dir: str) -> str:
        """
        Return user dir

        :param dir: dir name
        :return: user dir
        """
        if dir not in self.dirs:
            raise Exception('Unknown dir: {}'.format(dir))

        dir_data_allowed = ["img", "capture", "upload"]

        if self.has("upload.data_dir") \
                and self.get("upload.data_dir") \
                and dir in dir_data_allowed:
            path = os.path.join(self.get_user_path(), self.dirs["data"], self.dirs[dir])
        else:
            path = os.path.join(self.get_user_path(), self.dirs[dir])
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        return path

    def get_app_path(self) -> str:
        """
        Return app data path

        :return: app root path
        """
        if self.is_compiled():  # if compiled with pyinstaller
            return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        else:
            return os.path.abspath(os.path.dirname(__file__))

    def get_user_path(self) -> str:
        """
        Return user home path (workdir)

        :return: user home path
        """
        return self.path

    def init(self, all: bool = True):
        """
        Initialize config

        :param all: load all configs
        """
        if not self.initialized:

            # if app initialization
            if all:
                v = self.get_version()
                build = self.get_build()
                os = self.window.core.platforms.get_os()
                architecture = self.window.core.platforms.get_architecture()
                print("===================================================")
                print(" PyGPT    {} build {} ({}, {})".format(v, build.replace('.', '-'), os, architecture))
                print(" Author:  Marcin Szczyglinski")
                print(" GitHub:  https://github.com/szczyglis-dev/py-gpt")
                print(" Website: https://pygpt.net")
                print(" Email:   info@pygpt.net")
                print("===================================================")
                print("")
                print("Initializing...")

                # prepare and install
                self.window.core.installer.install()

            self.load(all)
            self.initialized = True

    def get_version(self) -> str:
        """
        Return version

        :return: version string
        """
        path = os.path.abspath(os.path.join(self.get_app_path(), '__init__.py'))
        try:
            with open(path, 'r', encoding="utf-8") as f:
                data = f.read()
                result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format("__version__"), data)
                return result.group(1)
        except Exception as e:
            if self.window is not None:
                self.window.core.debug.log(e)
            else:
                print("Error loading version file: {}".format(e))

    def get_build(self) -> str:
        """
        Return build

        :return: build string
        """
        path = os.path.abspath(os.path.join(self.get_app_path(), '__init__.py'))
        try:
            with open(path, 'r', encoding="utf-8") as f:
                data = f.read()
                result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format("__build__"), data)
                return result.group(1)
        except Exception as e:
            if self.window is not None:
                self.window.core.debug.log(e)
            else:
                print("Error loading version file: {}".format(e))

    def get_options(self) -> dict:
        """
        Return settings options

        :return: settings options
        """
        return self.provider.get_options()

    def get_sections(self) -> dict:
        """
        Return settings sections

        :return: settings sections
        """
        return self.provider.get_sections()

    def get(self, key: str, default: any = None) -> any:
        """
        Return config value by key

        :param key: key
        :param default: default value
        :return: value
        """
        if key in self.data:
            return self.data[key]
        return default

    def get_session(self, key: str, default: any = None) -> any:
        """
        Return session config value by key

        :param key: key
        :param default: default value
        :return: value
        """
        if key in self.data_session:
            return self.data_session[key]
        return default

    def get_lang(self) -> str:
        """
        Return language code

        :return: language code
        """
        test_lang = os.environ.get('TEST_LANGUAGE')  # if pytest
        if test_lang:
            return test_lang
        return self.get('lang', 'en')

    def set(self, key: str, value: any):
        """
        Set config value

        :param key: key
        :param value: value
        """
        self.data[key] = value

    def set_session(self, key: str, value: any):
        """
        Set session config value

        :param key: key
        :param value: value
        """
        self.data_session[key] = value

    def has(self, key: str) -> bool:
        """
        Check if key exists in config

        :param key: key
        :return: True if exists
        """
        if key in self.data:
            return True
        return False

    def has_session(self, key: str) -> bool:
        """
        Check if key exists in session config

        :param key: key
        :return: True if exists
        """
        if key in self.data_session:
            return True
        return False

    def all(self) -> dict:
        """
        Return all config values

        :return: dict with all config values
        """
        return self.data

    def all_session(self) -> dict:
        """
        Return all session config values

        :return: dict with all config values
        """
        return self.data_session

    def get_available_langs(self) -> list:
        """
        Return list with available languages

        :return: list with available languages (user + app)
        """
        langs = []
        path = os.path.join(self.get_app_path(), 'data', 'locale')
        if os.path.exists(path):
            for file in os.listdir(path):
                if file.startswith('locale.') and file.endswith(".ini"):
                    lang_id = file.replace('locale.', '').replace('.ini', '')
                    if lang_id not in langs:
                        langs.append(lang_id)

        path = os.path.join(self.get_user_path(), 'locale')
        if os.path.exists(path):
            for file in os.listdir(path):
                if file.startswith('locale.') and file.endswith(".ini"):
                    lang_id = file.replace('locale.', '').replace('.ini', '')
                    if lang_id not in langs:
                        langs.append(lang_id)

        # sort by name
        langs.sort()

        # make English first
        if 'en' in langs:
            langs.remove('en')
            langs.insert(0, 'en')

        # make Polish second
        if 'pl' in langs:
            langs.remove('pl')
            langs.insert(1, 'pl')
        return langs

    def append_meta(self) -> dict:
        """
        Append meta data

        :return: meta data dict
        """
        return {
            'version': self.version,
            'app.version': self.version,
            'updated_at': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        }

    def load(self, all: bool = True):
        """
        Load config

        :param all: load all configs
        """
        self.load_config(all)

        if all:
            self.window.core.modes.load()
            self.window.core.models.load()
            self.window.core.presets.load()
            self.window.core.plugins.load_presets()

    def load_config(self, all: bool = True):
        """
        Load user config from JSON file

        :param all: load all configs
        """
        self.data = self.provider.load(all)
        if self.data is not None:
            self.data = dict(sorted(self.data.items(), key=lambda item: item[0]))  # sort by key

    def load_base_config(self):
        """
        Load app config from JSON file
        """
        self.data_base = self.provider.load_base()
        if self.data_base is not None:
            self.data_base = dict(sorted(self.data_base.items(), key=lambda item: item[0]))  # sort by key
            self.initialized_base = True

    def from_base_config(self):
        """
        Load app config from JSON file
        """
        if not self.initialized_base:
            self.load_base_config()
        self.data = copy.deepcopy(self.data_base)

    def get_last_used_dir(self) -> str:
        """
        Return last used directory

        :return: last used directory
        """
        last_dir = self.get_user_dir("data")
        if self.has("dialog.last_dir"):
            tmp_dir = self.get("dialog.last_dir")
            if os.path.isdir(tmp_dir):
                last_dir = tmp_dir
        return last_dir

    def set_last_used_dir(self, path: str):
        """
        Save last used directory

        :param path: path
        """
        self.set('dialog.last_dir', path)
        self.save()

    def get_base(self, option: str = None) -> any:
        """
        Return base config option or all base config

        :param option: option name
        :return: option value or all config
        """
        if not self.initialized_base:
            self.load_base_config()
        if option is None:
            return self.data_base
        elif option in self.data_base:
            return self.data_base[option]

    def replace_key(self, data: dict, key_from: str, key_to: str) -> dict:
        """
        Replace key in dict

        :param data: dict
        :param key_from: key to replace
        :param key_to: new key
        :return: dict with replaced keys
        """
        if key_from in data:
            data[key_to] = copy.deepcopy(data[key_from])
            del data[key_from]
        return data

    def setup_env(self):
        """Setup environment vars"""
        if "app.env" not in self.data or not isinstance(self.data["app.env"], list):
            return
        list_loaded = []
        for item in self.data["app.env"]:
            if item['name'] is None or item['name'] == "":
                continue
            try:
                value = str(item['value'].format(**self.all()))
                os.environ[item['name']] = value
                list_loaded.append(item['name'])
            except Exception as e:
                print("Error setting env var: {}".format(e))
        if list_loaded:
            print("Setting environment vars: {}".format(", ".join(list_loaded)))

    def save(self, filename: str = "config.json"):
        """
        Save config

        :param filename: filename
        """
        self.provider.save(self.data, filename)
